""" This module provides the authentication class. """
# pylint:disable=E1101,R0902,W0403,R0913,R0903
import json
import os
import tempfile
import time

from ...shared.vendors import requests
from ..viewer.Proxies import Proxies

from ..Config import Config
from ..Singleton import Singleton


TOKENWARNINGS = [
    "Token valid.",
    "No access token available.",
    "Refresh token and access token expired.",
    "Access token expired and refresh token was invalid.",
]

TOKENVALID = 0
NOACCESSTOKEN = 1
TOKENSEXPIRED = 2
REFRESHINVALID = 3


class APIAuth(Singleton):
    """This is the authentication class. It stores access tokens and communicates with 3DM servers
    to procure new ones. This class acts as a singleton, meaning that it can be passed around with a
    simple call `APIAuth()` once it has been instantiated."""

    def __init__(
        self,
        username=None,
        password=None,
        twofactor=None,
        client_id="yasara-plugin-v2",
        use_system_proxies=True,
        try_auto_login=False,
        export_auth_on_login=False,
    ):
        """Initialize the APIAuth class. This method will only be executed the first time APIAuth()
        is called.

        :param username: Username for 3DM
        :param password: password for 3DM
        :param client_id: The client ID used to request OAuth tokens. Defaults to 'yasara-plugin'.
        :param use_system_proxies: Use proxy addresses defined in the system to connect to the
                                   internet.
        :param try_auto_login: Attempt to log in automatically using a previous session. Defaults to
                               False.
        :param export_auth_on_login: Store login tokens when login is succesful. Defaults to False.
        """

        self.oauth_server = Config().api_urls["oauth"] + "/oauth/token"
        self.__access_token = None
        self.__refresh_token = None
        self.__access_expiry_time = -1
        self.__refresh_expiry_time = -1
        self.__username = username
        self.__password = password
        self.__twofactor = twofactor
        self.client_id = client_id

        # Retrieves the appropriate client secret based on the API being talked to and client id.
        if "canary" in self.oauth_server:
            self.client_id = self.client_id.replace("-v2", "-bc")

        self.client_secret = Config().client_secret[self.client_id]

        self.__export_auth_on_login = export_auth_on_login

        if use_system_proxies:
            self.proxies = Proxies.get_proxies()
        else:
            self.proxies = {}

        auto_login_success = False
        if try_auto_login:
            auto_login_success = self.auto_import_api_auth()

        if not auto_login_success and (
            self.__username is not None and self.__password is not None
        ):
            self.connect(self.__username, self.__password, self.__twofactor)

    def __del__(self):
        """Cleans"""

    def get_username(self):
        """Returns the current username."""
        try:
            return self.__username
        except AttributeError:
            return None

    def _get_temp_path(self):
        """Returns the path where authentication tokens for this instance should be stored."""
        tmp = tempfile.gettempdir()
        filename = "%s-api-auth.json" % self.client_id
        return os.path.join(tmp, filename)

    def export_api_auth(self):
        """Exports all the necessary details to log in in a JSON file."""
        data = {
            "__access_token": self.__access_token,
            "__refresh_token": self.__refresh_token,
            "__access_expiry_time": self.__access_expiry_time,
            "__refresh_expiry_time": self.__refresh_expiry_time,
            "__username": self.__username,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }
        path = self._get_temp_path()
        with open(path, "w") as json_file:
            json.dump(data, json_file)

    def import_api_auth(self, path):
        """Imports a JSON file (`path`) that contains tokens and other login details and sets them
        as instance attributes. The token validity is checked, and if needed tokens are refreshed.
        Returns True if the import was successful, returns False if not."""

        with open(path, "r") as json_file:
            try:
                data = json.load(json_file)
            except ValueError:
                return False

        for key, value in data.items():
            if key.startswith("__"):
                # Compensate for private attributes
                key = "_APIAuth" + key
            if hasattr(self, key):
                setattr(self, key, value)
        return self._check_token_validity() == TOKENVALID

    def auto_import_api_auth(self):
        """Automatically gets the path of the tokens file and attempts to import it. Returns False
        if unsuccesful at any stage."""
        path = self._get_temp_path()
        if os.path.exists(path):
            return self.import_api_auth(path)
        return False

    def set_attributes_from_response(self, token):
        """Sets instance attributes from a response."""
        self.__access_token = token["access_token"]
        self.__refresh_token = token["refresh_token"]
        # This is most likely the current time + 2 hours
        self.__access_expiry_time = int(time.time()) + token["expires_in"]
        # Refresh tokens are valid for 3 days.
        self.__refresh_expiry_time = int(time.time()) + (3 * 24 * 3600)

    def store_token_through_password(self, username, password, twofactor=None):
        """Attempt to retrieve an access token from the OAuth2 server.
        If it is succesful, stores the token for later access.

        :param username: Username
        :param password: Password
        :param twofactor: Two-factor authentication code
        :raises: AssertionError if the server does not respond with a 200 HTTP code. If the error
                 message contains 'twofactor', try submitting again with a 2FA token."""

        response = requests.post(
            self.oauth_server,
            data=dict(
                username=username,
                password=password,
                twofactor=twofactor,
                grant_type="password",
            ),
            auth=(self.client_id, self.client_secret),
            proxies=self.proxies,
            verify=Config().verify_https,
            timeout=10,
        )

        assert response.status_code == 200, response.text

        data = response.json()

        self.set_attributes_from_response(data)

    def store_token_through_refresh(self):
        """Attempts to refresh the access token using a refresh token.
        If it is succesful, store the renewed token for later access."""
        assert self.__refresh_token is not None, "Refresh token not present."

        response = requests.post(
            self.oauth_server,
            data=dict(refresh_token=self.__refresh_token, grant_type="refresh_token"),
            auth=(self.client_id, self.client_secret),
            verify=Config().verify_https,
            proxies=self.proxies,
        )
        assert response.status_code == 200, response.text
        data = response.json()

        self.set_attributes_from_response(data)

    def _check_token_validity(self):
        """Checks if the tokens that are currently stored are still valid. If the access token has
        expired, it will attempt to refresh it using the refresh tokens.

        :returns: An integer corresponding to a TOKENWARNING. 0 (TOKENVALID) means the access token
        currently stored is valid.
        """
        if self.__access_token is None:
            return NOACCESSTOKEN
        if time.time() < self.__access_expiry_time:
            return TOKENVALID

        if time.time() < self.__refresh_expiry_time:
            try:
                self.store_token_through_refresh()
                return TOKENVALID
            except AssertionError:
                return REFRESHINVALID

        return TOKENSEXPIRED  # Both tokens have expired.

    def connect(self, username, password, twofactor=None, force=False):
        """Connects to the API and attempts to retrieve an access token.

        :param username: Username
        :param password: Password
        :param twofactor: Two-factor authentication code
        :param force: Optional, defaults to False. If set to true, the login procedure will take
                      place even if valid access tokens have already been acquired. This prevents
                      multiple unnecessary connection events from taking place.
        :returns: True if connecting was successful and a token was stored. False if connection
                  was not successful.
        :raises: Raises an AssertionError if the credentials were bad. If the error message contains
                 'twofactor', try submitting again with a 2FA token.
        """
        self.__username = username
        self.__password = password

        token_validity_response = self._check_token_validity()
        token_is_valid = token_validity_response == TOKENVALID

        if not force and token_is_valid:
            # Prevent multiple connection attempts when an access token has already been acquired.
            return True

        self.store_token_through_password(username, password, twofactor)

        if self.__export_auth_on_login:
            self.export_api_auth()
        return True

    def connect_with_token(self, access_token, token_expiry=7200, refresh_token=None):
        """Connects to the API with only an access token. This call should not be combined with
        the normal `connect` call.

        :param access_token: A valid API access token.
        :param token_expiry: Optional, the token expiry time in seconds. Defaults to 7200.
        :param refresh_token: Optional, if this is added, a refresh token with 72 hours
                              before the expiry time is added.
        :returns: True if connecting was successful and a token was stored. False if connection
                  was not successful.
        """
        self.__access_token = access_token
        self.__access_expiry_time = int(time.time()) + token_expiry
        if refresh_token is not None:
            self.__refresh_token = refresh_token
            self.__refresh_expiry_time = int(time.time()) + (72 * 3600)
        return True

    def get_request_headers(self):
        """Returns authorization headers."""
        if self._check_token_validity() is not TOKENVALID:
            self.connect(self.__username, self.__password, force=True)
        return {"Authorization": "Bearer %s" % self.__access_token}
