""" This module genericizes interation with the Basket API, allowing for easy interaction with
baskets. """

# pylint: disable=R0913,R0902,E1101,W0403

import json
from .API import API
from ..Config import Config


class BasketsAPI(object):
    """Provides methods for interacting with the baskets API."""

    def __init__(self, auth, user_agent=None):
        self.api = API("basket", auth, user_agent=user_agent)
        self.api.base_url = Config().api_urls["basket"]

    def list_baskets(self, sfamid=None, familyid=1, shared_with=None, basket_type=None):
        """Lists all the baskets available to a user.

        :param sfamid: Optional, always used in conjunction with basket_type. Specify this
                       if you want to list shared baskets for a given SFAMID.
        :param familyid: Optional, defaults to 1 (Full dataset)
        :param shared_with: Optional, this is an integer that specifies a group that the basket is
                            shared with.
        :param basket_type: The type of basket you'd like to list, like 'hotspot'.
        :returns: A list of Basket objects.
        :raises: A ValueError if either sfamid or basket_type is specified, but not both."""
        if basket_type is None:
            raise ValueError("You must specify a basket type.")

        suffix = ["type", basket_type]
        params = {"sfamid": sfamid, "sharedwith": shared_with, "familyid": familyid}
        # Remove all undefined parameters.
        params = {key: value for key, value in params.items() if value is not None}
        baskets = self.api.baskets(suffix=suffix, **params).json()

        try:
            baskets = baskets["_embedded"]["baskets"]
        except KeyError:
            try:
                # New basket response format.
                baskets = baskets["content"]
            except KeyError:
                return []  # No baskets found.
        return [Basket(basket, self) for basket in baskets]

    def load_basket(self, basket_id):
        """Loads a single basket by its ID.

        :param basket_id: Integer, the ID of a basket as created by the baskets API.
        :returns: A `Basket` with the information loaded from the API.
        :raises: AssertionError if the server did not respond with valid basket information."""
        req = self.api.baskets(suffix=[str(basket_id)], projection="withitems")
        assert (
            req.status_code == 200
        ), "Could not load basket #%s, status was %s: %s" % (
            basket_id,
            req.status_code,
            req.content,
        )
        basket_dict = req.json()
        return Basket(basket_dict, self)

    def delete_basket(self, basket):
        """Deletes a single basket, uses the `Basket` object.

        :param basket: A Basket object.
        """
        self.delete_basket_native(basket.id)

    def delete_basket_native(self, basket_id):
        """Deletes a basket, uses raw data.

        :param basket_id: The ID of a basket that needs to be removed."""
        response = self.api.baskets(suffix=[str(basket_id)], request_method="DELETE")
        assert response.status_code == 204, "Item (%s) was not removed: %s" % (
            basket_id,
            response.content,
        )

    def save_basket(self, basket):
        """Stores a basket, uses the `Basket` object.

        :param basket: A Basket object."""

        contents = str(basket)
        self.save_basket_native(
            contents,
            basket.id,
            new_items=basket.new_items,
            removed_items=basket.removed_items,
            initial_items=basket.initial_items,
        )

        # Stores the new items in the basket (to synchonize IDs.)
        basket.items = self.api.baskets(suffix=[str(basket.id), "items"]).json()[
            "_embedded"
        ]["items"]

    def _create_new_items(self, new_items, basket_id):
        """Creates new items in a basket.

        :param new_items: A list of new items (dict) for the basket.
        :param basket_id: The ID of the basket the items belong to."""
        for item in new_items:
            if "basket" not in item:
                item["basket"] = "/".join(
                    [self.api.base_url, "baskets", str(basket_id)]
                )
            if "label" not in item or "value" not in item:
                raise ValueError(
                    "Attempted to create a new incomplete item: %s" % json.dumps(item)
                )

            response = self.api.items(
                data=json.dumps(item),
                custom_headers={"Content-Type": "application/json"},
                request_method="POST",
            )
            assert response.status_code == 201, "Item (%s) was not created: %s" % (
                item,
                response.content,
            )

    @staticmethod
    def _create_altered_item(i_item, c_item):
        """Creates an 'altered' item for _update_altered_items. This method compares the item
        in its initial state to its current state, and builds a new dictionary containing directives
        for the future item.

        :param i_item: Initial item
        :param c_item: Current item
        :returns: An altered item.
        """
        altered_item = {}
        for key, value in c_item.items():
            # Update and add items
            if key not in i_item or value != i_item[key]:
                altered_item[key] = value

        # Remove deleted item properties
        for key in set(i_item).difference(c_item):
            altered_item[key] = None

        return altered_item

    def _update_altered_items(self, initial_items, current_items, perform_update=True):
        """Updates altered items in a basket.

        :param initial_items: The items that were initially in the basket before altering them.
        :param current_items: The items in their current state.
        :param perform_update: Defaults to True. When True, each of the altered items is PATCHed to
                               the server. If False, the altered items are just returned.
        """
        altered_items = {}
        if isinstance(initial_items, str):
            initial_items = json.loads(initial_items)

        for i_item in initial_items:
            for c_item in current_items:
                if i_item["id"] == c_item.get("id") and i_item != c_item:
                    altered_items[i_item["id"]] = self._create_altered_item(
                        i_item, c_item
                    )

        if not perform_update:
            return altered_items

        for _id, item in altered_items.items():
            response = self.api.items(
                suffix=[str(_id)],
                data=json.dumps(item),
                request_method="PATCH",
                headers={"Content-Type": "application/json"},
            )
            assert response.status_code == 200, "Item (%s) was not saved: %s" % (
                item,
                response.content,
            )
        return altered_items

    def _delete_removed_items(self, removed_items):
        """Removes items that were specified as being removed by the user.

        :param removed_items: A list of item IDs (int) of items that were removed.
        """
        for item_id in removed_items:
            response = self.api.items(suffix=[str(item_id)], request_method="DELETE")
            assert response.status_code == 204, "Item %s was not deleted: %s" % (
                item_id,
                response.content,
            )

    def save_basket_native(
        self,
        contents,
        basket_id,
        new_items=None,
        removed_items=None,
        initial_items=None,
    ):
        """Stores a basket, uses raw data.

        :param contents: The JSON representation of a basket, as produced by the API.
        :param basket_id: The ID of the basket.
        :raises: AssertionError if the basket was not saved properly."""

        # Set variables to empty lists if they equal None to prevent dangerous default values.
        new_items = [] if new_items is None else new_items
        removed_items = [] if removed_items is None else removed_items
        initial_items = [] if initial_items is None else initial_items

        self._create_new_items(new_items, basket_id)

        current_items = json.loads(contents)["items"]
        if isinstance(current_items, dict):
            current_items = current_items.values()

        self._update_altered_items(initial_items, current_items)

        self._delete_removed_items(removed_items)

    def new_basket(
        self,
        label,
        sfamid=None,
        basket_type="hotspot",
        items=None,
        familyid=1,
        metadata=None,
        shared_with=None,
    ):
        """Creates a new basket object.

        :param label: The label of the basket.
        :param sfamid: Optional, adds an sfamid to the basket, making it shared across that SFAM.
        :param basket_type: Optional, the type of basket. Defaults to "hotspot".
        :param items: Any items you'd like to add to the basket at the start. NOTE: The basket will
                      be created empty. It will still need to be saved with Basket.save() if you
                      specified items.
        :param familyid: The ID of the family this basket belongs to. The family is relevant
                         because alignment numbers can change between families. Defaults to 1.
        :param shared_with: Optional, the number of the group this basket is shared with. If this is
                            left out the default sharing behaviour (sharing with everyone who has
                            access to the system) is used by the API.
        :returns: A new Basket object.
        :raises: AssertionError if the basket could not be created.
        """
        if len(label) < 4:
            raise RuntimeWarning(
                "The label for a basket must be at least 4 characters long."
            )

        basket_basics = {
            "label": label,
            "sfamid": sfamid,
            "type": basket_type,
            "familyid": familyid,
            "metadata": json.dumps(metadata),
            "items": [],
        }
        if shared_with is not None:
            basket_basics["sharedwith"] = int(shared_with)

        req = self.api.baskets(
            data=json.dumps(basket_basics),
            request_method="POST",
            custom_headers={"Content-Type": "application/json"},
        )
        assert req.status_code == 201, "Could not create new basket: %s" % req.content

        basket = Basket(req.json(), basket_api=self)
        if items is not None:
            for item in items:
                basket.add_item(item)

        return basket


class Basket(object):
    """A Basket object is the Python representation of a basket."""

    def __init__(self, basket, basket_api=None):
        """Initializes the basket object using data from the server."""
        self._full_contents = basket
        self.basket_api = basket_api
        self.new_items = []
        self.removed_items = []
        self.metadata = ""
        for key, value in basket.items():
            setattr(self, key, value)

        if not hasattr(self, "id"):
            raise ValueError("Basket object must have an ID.")

        if not hasattr(self, "label"):
            self.label = None

        if not hasattr(self, "items"):  # The items list is vital.
            self.items = []

        # Store the initial items as a string, because strings are immutable.
        self.initial_items = json.dumps(basket.get("items", []))

    def __str__(self):
        """Returns a JSON string version of the basket with updated items."""
        self._full_contents["items"] = self.items  # Update items
        string = json.dumps(self._full_contents, indent=4)
        return string

    def add_item(self, item):
        """Adds an item to the items list."""
        self.items.append(item)
        self.new_items.append(item)

    def remove_item(self, item=None, where=None):
        """Removes an item from the items list.

        :param item: Optional, a specific item to delete.
        :param where: Optional, dict, a specific condition on which items should be removed.
                      For example {"value":1} deletes all items where `value` == 1.
        :raises: ValueError if neither item nor where is specified.
        """
        if item is not None:
            self.items.remove(item)
            try:
                self.removed_items.append(item["id"])
            except KeyError:
                pass  # A removed item without an ID is not yet present on the server.

            return
        if where is None:
            raise RuntimeWarning("Didn't specify anything to delete.")

        if not isinstance(where, dict):
            raise ValueError("The `where` parameter needs to be a dictionary.")
        # Copy the list when running through it, since we might alter it.
        for item in self.items[:]:  # pylint: disable=R1704
            for key, value in where.items():
                if item.get(key) == value:
                    self.items.remove(item)
                    try:
                        self.removed_items.append(item["id"])
                    except KeyError:
                        pass  # A removed item without an ID is not yet present on the server.

    def add_residue_position(self, position, comment=""):
        """Adds a residue position as an item."""
        self.add_item({"label": comment, "value": position})

    def save(self):
        """Saves the basket."""
        if self.basket_api is None:
            raise RuntimeError(
                "This basket was instantiated without an API reference. It cannot be\
            saved."
            )
        self.basket_api.save_basket(self)
        self.new_items = []
        self.removed_items = []
        self.initial_items = json.dumps(self.items)

    def delete(self):
        """Deletes the basket."""
        if self.basket_api is None:
            raise RuntimeError(
                "This basket was instantiated without an API reference. It cannot be deleted."
            )
        self.basket_api.delete_basket(self)
