""" This module facilitates the use of proxies with the molviewer-plugin. Using proxies in the past
has led to significant issues, and this class intends to provide methods that manage and test
proxy settings. """
# pylint: disable=C0103,E1101

import os
import json
import warnings
from ...shared.vendors import requests
from ...shared.Config import Config


PROXY_FILE_PATH = ".PROXIES"


class Proxies(object):
    """This class provides methods for retrieving, storing, managing and testing proxies. It deals
    mainly in `proxies` dictionaries, which should be formatted like this:

    {"http": <addr>, "https": <addr>}

    As per the requests standard (http://docs.python-requests.org/en/master/user/advanced/#proxies)
    """

    def __init__(self):
        pass

    @staticmethod
    def check_url(url):
        """This method checks if a URL is valid. Because URL validity checking is quite hard,
        especially when local top-level domains are taken into account, we are just checking for
        special characters that should not be allowed, and the correct protocol."""
        if not url.startswith("http"):
            return False
        for char in r"$^*;\!":
            if char in url:
                return False

        return True

    @staticmethod
    def get_proxies_from_file():
        """Returns the proxies based on the contents of the `PROXY_FILE_PATH`."""
        try:
            with open(PROXY_FILE_PATH) as f:
                return json.load(f)
        except IOError:
            warnings.warn("No proxies file exists. (%s)" % PROXY_FILE_PATH)
            return {}

    @staticmethod
    def test_http_proxy_connects(http_proxy_url):
        """Tests if the HTTP proxy works."""
        HTTP_URL = "http://api.bio-prodict.nl"
        try:
            r = requests.get(
                HTTP_URL,
                proxies={"http": http_proxy_url, "https": "https://invalid_url"},
                allow_redirects=False,
                timeout=30,
                verify=Config().verify_https,
            )
        except Exception as e:
            raise Exception(
                "Encountered the following error when connecting: %s. Please try checking your "
                "settings, or contact info@bio-prodict.nl if the problem persists."
                % e.message
            )
        if r.status_code == 403:
            return True
        elif r.status_code == 200:
            try:
                r.json()
            except Exception:
                raise Exception(
                    "Testing HTTP proxy connection failed (test_http_proxy_connects). \n"
                    "Connecting with the HTTP proxy did not yield the expected response.\n"
                    "Please check your proxy settings and contact info@bio-prodict.nl with this error "
                    "message if the problem persists. \n"
                    "Response was: %s" % r.content
                )
        else:
            raise Exception(
                "Testing HTTP proxy connection failed (test_http_proxy_connects). \n"
                "Connecting with the HTTP proxy did not yield the expected response. \n"
                "Please check your proxy settings and contact info@bio-prodict.nl with this error "
                "message if the problem persists. \n"
                "Response code was: %s" % r.status_code
            )

    @staticmethod
    def test_https_proxy_connects(https_proxy_url):
        """Tests if the HTTPS proxy works."""
        HTTPS_URL = "https://api.bio-prodict.nl"
        r = requests.get(
            HTTPS_URL,
            proxies={"https": https_proxy_url, "http": "http://invalid_url"},
            allow_redirects=False,
            timeout=30,
            verify=Config().verify_https,
        )
        if r.status_code == 200:
            return True
        else:
            raise Exception(
                "Testing HTTPS proxy connection failed (test_https_proxy_connects). \n"
                "Connecting with the HTTPS proxy did not yield the expected response.\n"
                "Please check your proxy settings and contact info@bio-prodict.nl with this error "
                "message if the problem persists. \n"
                "Something went wrong with the HTTPS proxy; response code was: %s"
                % r.status_code
            )

    @staticmethod
    def test_https_proxy_response(https_proxy_url):
        """Tests if the HTTPS proxy allows the correct response through."""
        HTTPS_URL = "https://api.bio-prodict.nl"
        r = requests.get(
            HTTPS_URL,
            proxies={"https": https_proxy_url},
            allow_redirects=False,
            timeout=30,
            verify=Config().verify_https,
        )
        try:
            data = r.json()
        except Exception:
            raise ValueError(
                "Testing HTTPS proxy response failed (test_https_proxy_response). \n"
                "Interacting with the HTTPS proxy did not yield the expected response.\n"
                "The server did not return JSON. This suggests that while your proxy was set"
                "correctly, our response did not come through correctly.\n"
                "Please contact info@bio-prodict.nl with this error message."
            )

        if "_links" in data.keys():
            return True

        raise ValueError(
            "Testing HTTPS proxy response failed (test_https_proxy_response). \n"
            "Interacting with the HTTPS proxy did not yield the expected response.\n"
            "The server did not return the expected JSON. This suggests that while your proxy was"
            "set correctly, our response did not come through correctly.\n"
            "Please contact info@bio-prodict.nl with this error message."
        )

    @staticmethod
    def test_proxies(proxies):
        """Tests the proxies provided by the proxy dictionary. This should always contain a HTTP
        and an HTTPS proxy."""
        pattern_error_msg = (
            "Your %s proxy did not match the proper URL pattern we were expecting. "
        )
        pattern_error_msg += "Please check the url for errors."

        if proxies["http"]:
            assert Proxies.check_url(proxies["http"]), pattern_error_msg % "HTTP"
        if proxies["https"]:
            assert Proxies.check_url(proxies["https"]), pattern_error_msg % "HTTPS"

        Proxies.test_http_proxy_connects(proxies["http"])
        Proxies.test_https_proxy_connects(proxies["https"])
        Proxies.test_https_proxy_response(proxies["https"])

        return True

    @staticmethod
    def dump_proxies_to_file(proxies):
        """Dumps a proxies dictionary to the `PROXY_FILE_PATH`."""
        assert "http" in proxies
        assert "https" in proxies
        with open(PROXY_FILE_PATH, "w") as f:
            json.dump(proxies, f, indent=4)

    @staticmethod
    def remove_proxies_file():
        """Removes the file containing proxies."""
        os.remove(PROXY_FILE_PATH)

    @staticmethod
    def get_proxies_from_environment():
        """Use the default lowercase environment variables for proxies to get a proper proxy
        parameter. Requests does use environment variables to find proxies, but the ones they
        specify are all uppercase (http://docs.python-requests.org/en/master/user/advanced/#proxies)
        and Linux environment variables are case-sensitive."""
        proxies = {}
        http_proxy = os.getenv("http_proxy", None)
        https_proxy = os.getenv("https_proxy", None)
        if http_proxy is not None and http_proxy != "":
            proxies["http"] = http_proxy
        if https_proxy is not None and https_proxy != "":
            proxies["https"] = https_proxy
        return proxies

    @staticmethod
    def get_proxies():
        """Attempts to retrieve proxies in any way possible. Prioritizes environment settings over
        those set in the `PROXY_FILE_PATH`."""
        proxies = Proxies.get_proxies_from_environment()
        if proxies != {}:
            return proxies

        proxies = Proxies.get_proxies_from_file()
        return proxies
