""" This is the Viewer class for YASARA. It implements the interaction with the host viewer. """
# pylint: disable=R0913,W0632,C0302,R0904,R0914,R0911,C0103,W0640,E1121
import os
import sys
import re
import json
import time
import base64
import warnings
import tempfile
import textwrap
from datetime import date

try:
    import cPickle
except ImportError:
    import pickle as cPickle

# Relative imports are used to support renaming the plugin directory inside YASARA's plg folder.
from .YasaraConsole import Console
from ...shared.vendors import six
from ...shared.Config import Config
from ...shared.api.API import API
from ...shared.vendors import requests
from ...shared.api.APIAuth import APIAuth
from ...shared.api.Baskets import BasketsAPI
from ...shared.viewer.ThreeDMWebsite import ThreeDMWebsite
from ...shared.viewer.Updates import Updates
from ...shared.viewer.CompositeUtilities import CompositeUtilities
from ...shared.viewer.Selection import Selection, ObjectIdentifier
from ...shared.viewer.Proxies import Proxies
from ... import __version__ as VERSION  # pylint: disable=E0611,W0406


# Make testing easier by automatically filling in the Username and Password in a testing environment
try:
    from ...shared.test.credentials import API_USERNAME, API_PASSWORD
except ImportError:
    API_USERNAME, API_PASSWORD = None, None

try:
    import yasara
    from container import container
except ImportError:
    if "sphinx" in sys.argv[0]:
        # Sphinx needs to be able to run this file to create documentation.
        pass
    else:
        raise RuntimeError("This viewer needs to be run from the yasara/plg directory.")


class YasaraViewer(object):
    """This is the Viewer class for YASARA. It implements the interaction with the host viewer."""

    def __init__(self):
        self._index_requestable_methods()
        self.storage = self._get_yasara_storage()

    @staticmethod
    def _get_yasara_storage():
        """Returns the YASARA storage container, creating it if it isn't available yet."""
        if yasara.storage is None:
            yasara.storage = container()
        return yasara.storage

    @staticmethod
    def has_valid_suffix(name):
        """Checks if an identifier has an accepted suffix."""
        return name[-5:] in ["_prot", "_drug", "_full"]

    def object_in_scene(self, obj):
        """Returns whether or not an object with the given name exists in the scene."""
        return self.structure_identifier(obj, "trunc") in yasara.NameObj("All")

    def name_object(self, query, new_name=None):
        """This is a wrapper for YASARA's NameObj. It automatically renames objects longer than 12
        characters with the `structure_identifier` method. It also automatically reconstructs full
        names from shortened identifiers."""
        if new_name is not None:
            if len(new_name) > 12:
                return yasara.NameObj(query, self.structure_identifier(new_name))
            else:
                return yasara.NameObj(query, new_name)

        return [
            self.structure_identifier(name, mode="full")
            for name in yasara.NameObj(query)
        ]

    def _rename_existing_truncated_structures(self, truncated):
        """If a structure name exists that has a conflicting truncated name, this method will
        rename the first entry to include a number."""
        full = self.storage.identifier_mapping[truncated]
        del self.storage.identifier_mapping[truncated]

        if self.has_valid_suffix(truncated):
            new_truncated = truncated[:6] + "1" + truncated[-5:]
        else:
            new_truncated = truncated[:11] + "1"

        self.storage.identifier_mapping[new_truncated] = full
        yasara.NameObj("Obj %s" % truncated, new_truncated)

    def _ordinal_truncate_with_suffix(self, truncated):
        """Properly truncates a truncated identifier if it has a valid suffix and other conflicting
        truncated identifiers exist. Returns None otherwise."""
        truncated_with_number_suffix = []
        for _id in self.storage.identifier_mapping:
            if (
                _id[:-6] == truncated[:-6]
                and _id[-6] in "0123456789"
                and _id[-5:] == truncated[-5:]
            ):
                truncated_with_number_suffix.append(_id)

        if truncated_with_number_suffix:
            if len(truncated_with_number_suffix) >= 9:
                raise RuntimeWarning(
                    "Too many objects with the same starting characters."
                )
            num = str(len(truncated_with_number_suffix) + 1)[:1]
            return truncated[:-6] + num + truncated[-5:]

    def _ordinal_truncate_no_suffix(self, truncated):
        """Properly truncates a truncated identifier if it does not have a valid suffix but other
        conflicting truncated identifiers exist. Returns None otherwise."""
        truncated_with_number_suffix = []
        for _id in self.storage.identifier_mapping:
            if _id[:11] == truncated[:11] and _id[-1] in "0123456789":
                truncated_with_number_suffix.append(_id)

        if truncated_with_number_suffix:
            if len(truncated_with_number_suffix) >= 9:
                raise RuntimeWarning(
                    "Too many objects with the same starting characters."
                )
            num = str(len(truncated_with_number_suffix) + 1)[:1]
            return truncated[:11] + num

    def _check_existing_truncated_structures(self, truncated):
        """This method checks if a given truncated structure name conflicts with any existing
        structure names. If it does not, the input name will be returned, otherwise it will return
        a unique structure name, preserving suffixes along the way. It may rename structures to
        make sure that structure names follow a logical path using the
        `_rename_existing_truncated_structures` method."""

        if truncated in self.storage.identifier_mapping.keys():
            self._rename_existing_truncated_structures(truncated)
            if self.has_valid_suffix(truncated):
                return truncated[:-6] + "2" + truncated[-5:]
            else:
                return truncated[:11] + "2"

        if self.has_valid_suffix(truncated):
            new_truncated = self._ordinal_truncate_with_suffix(truncated)
        else:
            new_truncated = self._ordinal_truncate_no_suffix(truncated)

        if new_truncated is not None:
            return new_truncated

        return truncated

    def _truncate_structure_identifier(self, name):
        """This method shortens a structure identifier so that it is 12 characters or less, which
        is the maximum length for a given structure name. It preserves _prot, _drug and _full
        suffixes."""
        if self.has_valid_suffix(name):
            suffix = name[-5:]
            truncated = name[:-5][:7] + suffix
        else:
            truncated = name[:12]

        # Make sure the final character is not a number
        truncated = truncated.rstrip("0123456789")

        truncated = self._check_existing_truncated_structures(truncated)
        return truncated

    def structure_identifier(self, identifier, mode="toggle"):
        """This method allows the YASARA plugin to deal with structures that have names longer than
        12 characters. Wrap any PDB or model id in the `structure_identifier` method and it will
        resolve to an appropriate version based on the input. This means that long identifiers
        will result in the truncated version being retrieved, while inserting a (known) truncated
        name will result in the proper full name being returned, making this method a seamless
        toggle. Duplicate names will also be made unique.

        TLDR: Got a long name and it needs to be short? Use `structure_identifier()`. Got a short
        name and it needs to be long? Use `structure_identifier()`.

        The `identifier_mapping` storage object contains a mapping of structure identifiers:

            {truncated_name: full_name}

        :param identifier: An arbitrary identifier you want to convert. Also accepts Selections with
                           object names.
        :param mode: Conversion mode. Defaults to "toggle", which automatically switches between
                     long and short modes. When "full" is specified, the full identifier will always
                     be returned (when available). When "trunc" is specified, the truncated version
                     of the identifier will always be returned.
        """
        if not any(
            [isinstance(identifier, _class) for _class in [six.string_types, Selection]]
        ):
            return identifier

        if isinstance(identifier, Selection):
            # Handle object names in Selection objects, reassign object names with correct mode.
            selection_instance = identifier
            selection_instance.not_in_object = self.structure_identifier(
                selection_instance.not_in_object, mode=mode
            )
            selection_instance.in_object = self.structure_identifier(
                selection_instance.in_object, mode=mode
            )
            return selection_instance

        try:
            self.storage.identifier_mapping
        except AttributeError:
            self.storage.identifier_mapping = {}

        identifier_mapping = self.storage.identifier_mapping  # Truncated -> full
        identifier_mapping_inv = {
            v: k for k, v in identifier_mapping.items()
        }  # Full -> truncated

        if mode in ["toggle", "trunc"]:
            if (
                identifier in identifier_mapping_inv
            ):  # Check if this is the full name for a id
                return identifier_mapping_inv[
                    identifier
                ]  # Returns the truncated identifier

            if len(identifier) > 12:
                # This is a new long identifier
                truncated = self._truncate_structure_identifier(identifier)
                self.storage.identifier_mapping[truncated] = identifier
                return truncated

        if mode in ["toggle", "full"]:
            # From here on out the identifier *will* be a truncated version.
            if identifier in identifier_mapping:
                return identifier_mapping[identifier]

            # Could've been renamed
            if identifier[:11] + "1" in identifier_mapping:
                return identifier_mapping[identifier[:11] + "1"]

            # Could be a rename with a suffix
            if identifier[:6] + "1" + identifier[-5:] in identifier_mapping:
                return identifier_mapping[identifier[:6] + "1" + identifier[-5:]]

        if identifier not in identifier_mapping:
            identifier_mapping[identifier] = identifier

        return identifier

    def duplicate_view(self, name):
        """This method duplicates the view in YASARA, whilst keeping track of all the views that
        have been created up to that point. It will add a number if a view was already created at
        some point, preventing "View already exists" errors."""

        name = CompositeUtilities.get_valid_view_name(self, name)
        try:
            yasara.DuplicateView("Main", name=name, hud="Yes")
        except RuntimeError as e:
            raise RuntimeWarning(str(e))

        # Make sure that the secondary structure is showing to prevent edge cases with disappearing
        # residues when secondary structure is hidden in the Main tab.
        yasara.ShowSecStrAll("Tube")
        return name

    def check_structures_exist_in_scene(self):
        """Checks if there are any structures in the scene at all. If not, raise a RuntimeWarning
        that the user needs to load an object into the scene."""

        if not yasara.CompoundMol("All"):
            raise RuntimeWarning(
                "No objects in the scene. Please load a structure from 3DM."
            )

    def check_scene_objects(self):
        """Checks the scene if objects loaded by 3DM are present. If so, adjusts the system/family
        and subset identifiers such that they correspond to that object.

        :returns:  True if a 3DM object was found."""
        descriptions = yasara.CompoundMol("All")

        # No objects in the scene
        if not descriptions:
            return False

        found_3dm_object = False
        changed_system = False
        for desc in descriptions:
            split_desc = desc.split(";")
            if len(split_desc) != 7:
                continue

            # Find and rename the object to the correct truncated name.
            obj_name = self.find_obj_name_for_compound_mol(desc)
            if obj_name is None:
                # The object has already been renamed.
                print("Could not find object name for CompoundMol string: %s" % desc)
                continue

            found_3dm_object = True

            try:
                if obj_name in self.storage.numbering_schemes:
                    # No need to rescan this if it was already registered
                    continue
            except AttributeError:
                # Numbering schemes are not defined.
                self.storage.numbering_schemes = {}

            new_sfamid = split_desc[0].split(" ")[-1]

            if hasattr(self.storage, "system") and new_sfamid != self.storage.system:
                changed_system = True

            self.storage.system = new_sfamid
            self.storage.system_nice_name = (
                CompositeUtilities.get_nice_name_for_system_id(self, new_sfamid)
            )

            self.storage.family = split_desc[1]
            self.storage.subset = int(split_desc[2])
            self.storage.all_subsets = CompositeUtilities.load_alignment_subsets(
                self, sfamid=self.storage.system
            )

            # Find the full object name in the CompoundMol str and its corresponding truncated name.
            full_obj_name = split_desc[-2]

            # Prevent protein/compound naming convention from failing
            if "_prot" in obj_name and not full_obj_name.endswith("_prot"):
                full_obj_name += "_prot"
            elif "_drug" in obj_name and not full_obj_name.endswith("_drug"):
                full_obj_name += "_drug"
            elif "_full" in obj_name and not full_obj_name.endswith("_full"):
                full_obj_name += "_full"

            # Change compound mol string
            new_compound_mol_string = list(split_desc)
            new_compound_mol_string[-2] = full_obj_name
            new_compound_mol_string = ";".join(new_compound_mol_string)
            yasara.CompoundMol("Obj %s" % obj_name, new_compound_mol_string)

            trunc_obj_name = self.structure_identifier(full_obj_name, "trunc")

            yasara.NameObj(obj_name, trunc_obj_name)

            # Save the correct name in storage to keep track of it with the numbering scheme
            if full_obj_name not in self.storage.numbering_schemes:
                self.storage.numbering_schemes[full_obj_name] = split_desc[-1]

        if found_3dm_object and not self.storage.numbering_schemes:
            raise RuntimeError(
                "3DM objects found, but none are stored in the numbering schemes storage."
            )

        if found_3dm_object and changed_system:
            yasara.ShowMessage(
                "You are now using: %s - %s"
                % (self.storage.system, self.storage.family)
            )

        return found_3dm_object

    def clear_scene(self, certain=False):
        """Completely clears the scene and any lingering storage. Asks the user for permission."""
        if not certain:
            confirm = yasara.ShowWin(
                "Custom",
                "Are you sure you want to clear the scene?",
                500,
                182,
                "TextCen",
                250,
                48,
                "Are you sure you want to clear the scene and storage?",
                "TextCen",
                250,
                70,
                "This will remove all the objects on screen and reset the",
                "TextCen",
                250,
                92,
                "system and alignment settings.",
                "Button",
                52,
                130,
                "Clear",
                "Button",
                440,
                130,
                "Go back",
            )
            certain = confirm[0] == "Clear"
        # This if is not redundant, see here ^
        if certain:
            yasara.Clear()
            yasara.storage = container()
            self.storage = self._get_yasara_storage()

    def _store_object(self, name, obj):
        """Stores an object in YASARA storage as a pickled string. YASARA will sometimes crash
        when storing objects in the storage container. Using cPickle to convert any object to a
        string allows you to store arbitrary objects in YASARA for later retrieval using
        `_get_stored_object`.

        :param name: The name of the object you are storing. If this is not unique, the previous
                     value will be overwritten.
        :param obj: The Object you are storing. Anything that can be pickled can be used here."""
        setattr(self.storage, name, cPickle.dumps(obj))

    def _get_stored_object(self, name):
        """Returns the object stored under `name` in its native form. Returns `None` if the object
        does not exist."""
        try:
            return cPickle.loads(getattr(self.storage, name))
        except AttributeError:
            return None

    def _index_requestable_methods(self):
        """Runs through all the attributes in this class, checks if they are callable and don't
        start with an underscore (_). Methods that adhere to these prerequisites are added to the
        self.requestableMethods dictionary and can be called using _runYasaraRequest."""
        self.requestable_methods = {}
        for attribute_name in dir(self):
            attribute = getattr(self, attribute_name)
            # Check if the attribute is a function and does not start with an underscore.
            if hasattr(attribute, "__call__") and attribute_name[0] != "_":
                self.requestable_methods[attribute_name] = attribute

    @staticmethod
    def _unknown_request(request=None):
        """This method is run automatically if _runYasaraRequest is called with a request string
        that does not correspond to an existing method in this class.
        This method can be used to gracefully handle any errors that may cause.

        :returns: False
        """
        warnings.warn(
            "The 3DM plugin was requested to perform an unknown method: '%s'" % request
        )
        return False

    def _run_yasara_request(self, request):
        """This function is run with the `yasara.request` string as an argument. This saves
        developers from having to add to/search through a massive if/if/if block at the bottom
        of the plugin.

        This will also deal with requests for showing superfamily data.

        :param request: The yasara.request string.
        :returns: The output of the requested method, or False if an unknown request was made.
        """
        # Remove all existing messages from screen
        yasara.HideMessage()

        # Notify user about updates if available and allowed
        version_available = Updates.new_version_available(VERSION, "yasara")
        if version_available:
            self.show_update_notification(version_available)

        # Run the proper method, if it is defined. If not, fall back to self._unknownRequest.
        try:
            # Most of the time a request will fit here.
            requested_method = self.requestable_methods[request]
        except KeyError:
            if request.startswith("show_superfamily_data_"):
                # Ex: show_superfamily_data_mutations will yield show_superfamily_data("mutations")
                return self.show_superfamily_data(request.split("_")[-1])
            if request.startswith("color_molecules_by_"):
                return self.color_molecules_by(request.split("_")[-1])
            if request.startswith("show_literature_hotspots_kw"):
                # Call literature hotspots with predefined keywords
                return self.show_literature_hotspots(
                    keywords=request.split("_")[4:], cutoff=2
                )
            if request == "__init__":
                # Ignore requested __init__ call
                return

            # Fall back to unknown request warning
            return self._unknown_request(request)

        return requested_method()

    def _get_api(self, api="persistence3dmdata"):
        """Tries to return the API() object by accessing the stored API authentication instance.
        If it does not exist, attempts to log in the user.

        :param api: The Bio-Prodict API one is trying to access. Defaults to `persistence3dmdata`.
        :returns: An authenticated API instance."""
        self._check_user_access()
        return API(api, self.storage.api_auth, user_agent=self.get_version())

    def _get_baskets_api(self):
        """Returns a Baskets API interaction object."""
        self._check_user_access()
        return BasketsAPI(self.storage.api_auth, user_agent=self.get_version())

    def _check_user_access(self):
        """Checks if the user is logged in to the API. If not, attempt to log in automatically.
        Should that not work, then the user is presented with a login prompt."""
        try:
            self.storage.api_auth  # pylint: disable=W0104
            return True
        except AttributeError:
            pass

        api_auth = APIAuth(try_auto_login=True, export_auth_on_login=True)

        if api_auth._check_token_validity() == 0:  # pylint: disable=W0212
            self.storage.api_auth = api_auth
            yasara.ShowMessage("Logging in as %s." % api_auth.get_username())
        else:
            if Config().use_scene_login:  # pylint: disable=E1101
                self.login_with_scene()
            else:
                self.api_login(prefill_username=api_auth.get_username())

    def api_login(self, retry=False, credentials=(), prefill_username=None):
        """Prompts the user for their username and password to log in to the API. If the login
        attempt is successful, storage the API object in YASARA storage.

        :param retry: Defaults to False. If this parameter is set to True, the Window will display
                      that the previous login attempt failed.
        :param credentials: Used in testing to bypass login window. Two-tuple containing a username
                            and password.
        :param prefill_username: A username that prefills the username field. Is overwritten in test
                                 scenarios.
        """
        if retry:
            message = "Your previous log in attempt was rejected. Please try again."
        else:
            message = "Please enter your 3DM username and password."

        if len(credentials) == 0:
            username_string = "Username"
            password_string = "Password"
            # Load automatically add username and password if these are available.
            if prefill_username is not None:
                username_string = "Username," + prefill_username

            if API_USERNAME is not None:
                username_string = "Username," + API_USERNAME
            if API_PASSWORD is not None:
                password_string = "Password," + API_PASSWORD

            credentials = yasara.ShowWin(
                "TextInput", "Login", message, username_string, password_string
            )

        if len(credentials) == 2:
            username, password = credentials
        else:
            raise RuntimeWarning(
                "Logging into 3DM canceled. Stopping. Can't continue without login."
            )

        try:
            self.storage.api_auth = APIAuth(
                username=username, password=password, export_auth_on_login=True
            )
            yasara.HideMessage()
        except AssertionError as ass_err:
            # AssertionError is raised if the server declined our API Authentication request
            # (wrong password/user), limit reached. It can also be raised if two factor auth is
            # required, which is handled in handle_twofactor.
            if len(credentials) == 0:
                self.api_login(retry=True)
            else:
                if "twofactor" in str(ass_err):
                    self.handle_twofactor(username, password)
                else:
                    try:
                        default_err_message = "Wrong username or password."
                        err = json.loads(str(ass_err)).get(
                            "error_description"
                        )
                        if ("Bad credentials" in err or default_err_message in err) and "?" in password:
                            err = (
                                "Wrong username or password. "
                                "Your password may contain non-standard characters, these are incompatible with YASARA. "
                                "Please consider changing your password."
                            )
                        raise RuntimeWarning(
                            err
                        )
                    except json.decoder.JSONDecodeError:
                        raise RuntimeWarning("Login error: " + str(ass_err))

    def handle_twofactor(self, username, password, retry=False):
        """Handles twofactor authentication in a separate window."""
        if retry:
            message = "Your previous log in attempt was rejected. Please try again."
        else:
            message = "Please enter your two-factor authentication token."

        twofactor = yasara.ShowWin(
            "TextInput", "Login", message, "Two factor authentication token"
        )
        try:
            twofactor = twofactor[0]
        except IndexError:
            self.handle_twofactor(username, password, retry=True)

        try:
            self.storage.api_auth = APIAuth(
                username=username,
                password=password,
                twofactor=twofactor,
                export_auth_on_login=True,
            )
            yasara.HideMessage()
        except AssertionError as ass_err:
            raise RuntimeWarning(
                json.loads(str(ass_err)).get(
                    "error_description", "Wrong username or password."
                )
            )

    def login_with_scene(self):
        """Attempts to authenticate the user by means of a generated YASARA scene."""
        if yasara.ListObj("All"):
            raise RuntimeWarning("Please make sure there are no loaded objects.")

        results = yasara.ShowWin(
            "Custom",
            "Authenticate with Bio-Prodict Scene",
            600,
            200,
            "TextCen",
            300,
            48,
            "Authenticate with Bio-Prodict Scene",
            "TextCen",
            300,
            68,
            "Use the button below to get your YASARA scene, ",
            "TextCen",
            300,
            88,
            "then open it in the next screen.",
            "Button",
            200,
            128,
            "_L_aunch",
            "Button",
            400,
            128,
            "_C_ancel",
        )
        if "Launch" in results:
            yasara.ShowURL(ThreeDMWebsite.get_login_scene_url("yasara"))
            f = yasara.ShowWin(
                "FileSelection", "Select your authentication scene.", "No", "*.sce"
            )
            try:
                yasara.LoadSce(f[-1])
            except IndexError:
                raise RuntimeWarning(
                    "No login scene provided. Authentication canceled."
                )
            self.extract_auth_details()

    def extract_auth_details(self):
        """Attempts to extract authentication details from the scene."""
        if yasara.ListObj("All") and len(yasara.ListObj("All")) > 1:
            raise RuntimeWarning("Please make sure you have loaded only a login scene.")

        try:
            data = base64.b64decode("".join(yasara.NameRes("All")))
            auth_details = json.loads(data)
        except ValueError:
            raise RuntimeWarning(
                "Invalid authentication scene. Please try downloading again."
            )

        self.storage.api_auth = APIAuth()
        self.storage.api_auth.connect_with_token(
            access_token=auth_details["AccessToken"],
            token_expiry=3600,
            refresh_token=auth_details["RefreshToken"],
        )
        self.storage.api_auth.export_api_auth()

        yasara.DelAll()  # Clears the authentication object
        yasara.ShowMessage("Successfully extracted login details. You can now use 3DM.")

    def load_pdb_batch(self, *args, **kwargs):
        """Used as a redirect for the CompositeUtilities function."""
        CompositeUtilities.load_pdb_batch(self, *args, **kwargs)

    def load_pdb(
        self,
        pdbid,
        sfamid=None,
        specifier="chain",
        api_endpoint=None,
        auto_renumber=True,
        *args,
        **kwargs
    ):
        """Downloads a PDB file from the 3DM API and loads it into YASARA.

        :param pdbid: The PDB ID.
        :param sfamid: The superfamily ID.
        :param specifier: The specifier argument for the API. This allows you to filter what parts
                          of the structure are to be retrieved. "chain" is the default, results in
                          the entire chain. "amino" only downloads the protein, without any ligands.
                          "hetatam" only loads ligands.
        :param api_endpoint: Optional, you can use this to use a custom api endpoint to load a PDB.
                             This argument should be a method of the API class that is logged in.
                             All other args and kwargs are pased to the api endpoint call.
        :param auto_renumber: Optional, whether or not the PDB should be automatically renumbered
                              to the align numbering scheme. Defaults to True.
        :returns: The name of the loaded object.
        """

        if sfamid is None:
            sfamid = CompositeUtilities.get_sfamid(self)

        abs_pdb_path = CompositeUtilities.download_pdb_file(
            self, pdbid, sfamid, specifier, api_endpoint=api_endpoint, *args, **kwargs
        )
        obj_name = self.load_pdb_file_into_viewer(
            abs_pdb_path, pdbid, sfamid, specifier, auto_renumber
        )
        return obj_name

    def load_pdb_file_into_viewer(
        self, abs_pdb_path, pdbid, sfamid, specifier="chain", auto_renumber=True
    ):
        """Loads a PDB file into YASARA from a path. Also takes care of registering and
        renumbering the object. Deletes the file when done.

        :param abs_pdb_path: An absolute path to the PDB file.
        :param pdbid: The PDB ID
        :param sfamid: The superfamily identifier.
        :param specifier: The specifier argument for the API. This allows you to filter what parts
                          of the structure are to be retrieved. "chain" is the default, results in
                          the entire chain. "amino" only downloads the protein, without any ligands.
                          "hetatam" only loads ligands. Finally, "full" loads the entire structure
                          contained in the PDB, including nucleotides.
        :param auto_renumber: Optional, whether or not the PDB should be automatically renumbered
                              to the align numbering scheme. Defaults to True.
        :returns: The name of the loaded object.
        """
        is_first_object = len(yasara.ListObj("All")) == 0
        
        yasara.LoadPDB(abs_pdb_path)
        default_obj_name = os.path.basename(abs_pdb_path)[:12].replace(
            ".", "_"
        )  # Get YASARA name

        obj_name = pdbid
        if specifier == "hetatm":
            obj_name += "_drug"
        elif specifier == "amino":
            obj_name += "_prot"
        elif specifier == "full":
            obj_name += "_full"

        # Rename the object to the PDB ID (plus suffix) instead of temp filename.
        self.name_object("Obj %s" % default_obj_name, obj_name)
        # yasara.DelFile(abs_pdb_path)  # Delete the file once YASARA is done with it.

        if is_first_object:
            # Set the scene display to the default tube style if this is the
            # first object.
            yasara.Run("Style Tube")

        # Add the default numbering scheme, that being PDB.
        try:
            self.storage.numbering_schemes[obj_name] = "pdb"
        except AttributeError:
            self.storage.numbering_schemes = {obj_name: "pdb"}

        # Renumber PDB to alignment numbering
        if auto_renumber:
            try:
                family = CompositeUtilities.get_familyid(self)
            except RuntimeWarning:
                family = 1

            self.renumber_pdb(obj_name, sfamid, family, "align")

        return obj_name

    def load_corresponding_compounds(self, download_mode="ligand", objects=None):
        """Downloads partial objects based on a list of objects set by the user, either through
        the optional parameters, or a dialog.

        :param download_mode: String, either protein or ligand. Corresponds to which item type
                              must be downloaded.
        :param objects: A list of PDB ids where the missing item should be downloaded."""
        if objects is None or not objects:
            download_mode, objects = self.show_corresponding_compounds_dialog()

        specifier = {"protein": "amino", "ligand": "hetatm", "full": "full"}[
            download_mode
        ]
        for obj in objects:
            # Make sure we only get the PDB ID
            _, pdb_id = CompositeUtilities.split_structure_accession(obj)
            pdb_id = pdb_id[:5]

            self.load_pdb(pdb_id, CompositeUtilities.get_sfamid(self), specifier)

        self.transfer_pdb_coordinates(objects)

    def show_corresponding_compounds_dialog(self):
        """Shows the user a dialog containing all the objects in the scene and presents them with
        the choice between downloading corresponding proteins or ligands. Returns the download model
        (string, either protein or ligand), and a list of PDB ids that were selected by the user."""

        pdbs = self.name_object("All")

        window_arguments = [
            "Custom",
            "Load corresponding compounds",
            600,
            400,
            "Button",
            550,
            348,
            "_O_ K",
            "RadioButtons",
            3,
            2,
            20,
            348,
            "Protein",
            150,
            348,
            "Compounds",
            310,
            348,
            "Full structure",
            "List",
            20,
            48,
            "Objects in scene",
            560,
            270,
            "yes",
            len(pdbs),
        ]
        window_arguments += pdbs
        options = yasara.ShowWin(*window_arguments)

        modes = {1: "protein", 2: "ligand", 3: "full"}
        try:
            download_mode = modes[options[0]]
        except (KeyError, IndexError):
            raise RuntimeWarning("Invalid download mode.")

        if options[1] == 0:
            raise RuntimeWarning(
                "You did not select any objects to match compounds with."
            )

        objects = [pdb.split("_")[0] for pdb in options[2:]]
        return download_mode, objects

    @staticmethod
    def transfer_pdb_coordinates(pdbids):
        """Transfers the local coordinates of each of the pdbs in the input to the coordinate
        system of the first available object in the workspace, or the first pdb in the list.
        This will superimpose all the pdbs in the input.

        :param pdbids: A list of PDB IDs (corresponding to object IDs.)
        """
        existing_objects = yasara.ListObj("Atom All")
        if existing_objects:
            initial_pdb = existing_objects[0]
            for obj_id in existing_objects[1:]:
                yasara.TransferObj(obj_id, initial_pdb)

        elif len(pdbids) > 1:
            for pdbid in pdbids[1:]:
                yasara.TransferObj(pdbid, pdbids[0])

    def renumber_pdb(self, *args, **kwargs):
        """Forwards the renumber_pdb method calls to CompositeUtilities."""
        return CompositeUtilities.renumber_pdb(self, *args, **kwargs)

    def renumber_object_to_resnum(self, objects=None, negative=False):
        """This method will iterate over all the residues in the specified objects and renumber
        them to residue numbering.

        :param objects: A list of ObjectIdentifiers in the scene that must be renumbered.
        :param negative: If set to True, renumbering will be negative (-1--N). Defaults to false."""

        if negative:
            first, increment = -1, -1
        else:
            first, increment = 1, 1

        if isinstance(objects, ObjectIdentifier):
            objects = [objects]

        for obj in objects:
            assert isinstance(obj, ObjectIdentifier)
            name = self.structure_identifier(obj.name, mode="trunc")
            if obj.chain:
                yasara.RenumberRes(
                    "all in Obj %s Mol %s" % (name, obj.chain),
                    first=first,
                    increment=increment,
                )
            else:
                yasara.RenumberRes(
                    "all in Obj %s" % name, first=first, increment=increment
                )

    def get_stable_numbering_selections(self, obj):
        """Returns a stable list of residue identifiers that do not change during renumbering."""
        obj_name = self.structure_identifier(obj.name, mode="trunc")
        if obj.chain:
            return yasara.ListRes("Protein in Obj %s Mol %s" % (obj_name, obj.chain))
        else:
            return yasara.ListRes("Protein in Obj %s" % obj_name)

    def renumber_residue(self, _from, _to, extension=None):
        """Renumbers a single residue.

        :param _from: Either a Selection object (which will be used directly) or a residue number
                      (which will be inserted into a PyMol selector)
        :param _to: The target number for the given residue.
        :param extension: A character that is added as a suffix to the residue number.
        """
        if isinstance(_from, Selection):
            _from = self.structure_identifier(_from, "trunc")
            yasara.RenumberRes(_from.yasara(), first=_to, inscode=extension)
        if isinstance(_from, six.string_types):
            yasara.RenumberRes(_from, first=_to, inscode=extension)
        else:
            yasara.RenumberRes(
                Selection(numbers=_from).yasara(), first=_to, inscode=extension
            )

    def _numbering_refresh(self, new_family_id=None, new_subset_id=None):
        """Renumbering must take place when switching families or subsets if objects are numbered
        using the alignment numbering scheme.

        :param new_family_id: The ID of the family that the new numbering should use. If this is not
                              set, use the current family id.
        :param new_subset_id: The ID of the subset that the new numbering should use. If this is not
                              set, use the current subset id."""
        if new_family_id is None:
            new_family_id = CompositeUtilities.get_familyid(self)
        if new_subset_id is None:
            new_subset_id = CompositeUtilities.get_subset_id(self)

        if not hasattr(self.storage, "numbering_schemes"):
            return  # No objects have been loaded.

        for pdbid, numbering in self.storage.numbering_schemes.items():
            if numbering != "align":
                continue

            # Renumber to residue numbering (which is the same across families) and back to align
            # with a new numbering scheme.
            try:
                self.renumber_pdb(
                    pdbid,
                    CompositeUtilities.get_sfamid(self),
                    CompositeUtilities.get_familyid(self),
                    "res",
                )
                self.renumber_pdb(
                    pdbid, CompositeUtilities.get_sfamid(self), new_family_id, "align"
                )
            except ValueError as error:
                with Console(yasara):
                    print(error)

    def show_renumbering_dialog(self, custom_numbering_types=None):
        """Shows the user a dialog containing all the objects in the scene and all the
        numbering schemes they can convert to. Returns a target numbering scheme (string) and a list
        object names in YASARA. Shows a warning if the user didn't select any objects."""
        if custom_numbering_types is None:
            custom_numbering_types = []
        pdbs = self.name_object("All")

        if len(pdbs) == 0:
            raise RuntimeWarning(
                "You need to load an renumberable object from 3DM first."
            )

        # Due to the nature of having two dynamic lists, this more complex window_arguments
        # construction is required.
        window_arguments = [
            "Custom",
            "Renumber objects",
            600,
            400,
            "Button",
            550,
            348,
            "_O_ K",
            "List",
            230,
            48,
            "Renumber to...",
            352,
            270,
            "no",
            3 + (len(custom_numbering_types)),
        ]
        # List of renumbering options.
        window_arguments += [
            "3DM alignment numbering (align)",
            "PDB numbering (pdb)",
            "Residue numbering (res)",
        ]
        window_arguments += ["(%s)" % cnt for cnt in custom_numbering_types]
        # List of objects found in the environment
        window_arguments += ["List", 20, 48, "Objects", 200, 270, "yes", len(pdbs)]
        window_arguments += pdbs

        # Use the window_arguments list to show the window.
        options = yasara.ShowWin(*window_arguments)

        try:
            target_numbering = options[1].split("(", 1)[1][:-1]
            objects = options[3:]
        except IndexError:
            raise RuntimeWarning("Nothing was selected.")

        if len(objects) == 0:
            raise RuntimeWarning(
                "You did not select any objects to renumber. Please try again."
            )
        return target_numbering, objects

    def renumber_objects(self, target=None, objects=None):
        """Accepts a target numbering scheme and a list of objects to change to that numbering
        scheme. If either of these is not supplied, the user will be prompted for them. This
        method uses these values to renumber the given objects to the target scheme. The source
        scheme is automatically determined."""
        if target is None or objects is None:
            # Prompt the user to supply the objects and target renumbering
            custom_numbering_scheme_names = (
                CompositeUtilities.get_custom_numbering_schemes(self)
            )
            target, objects = self.show_renumbering_dialog(
                sorted(custom_numbering_scheme_names)
            )

        # Retrieve the numbering type id from the list string.
        for obj in objects:
            try:
                source = self.storage.numbering_schemes[obj]
            except (KeyError, AttributeError):
                # This value is only set if the object has already been renumbered.
                # If not, assume PDB numbering
                source = "pdb"

            # Renumbering for models
            accession, pdb = CompositeUtilities.split_structure_accession(obj)

            yasara.ShowMessage(
                "Busy renumbering %s from %s to %s..." % (pdb, source, target)
            )
            self.renumber_pdb(
                pdb,
                CompositeUtilities.get_sfamid(self),
                CompositeUtilities.get_familyid(self),
                target,
                accession=accession,
            )
            yasara.ShowMessage(
                "Renumbering %s from %s to %s complete." % (pdb, source, target)
            )

    @staticmethod
    def _filter_id_list(id_list, search):
        """Convert the search query to a regex, replace the * with a wildcard regex statement. Uses
        the search query to filter through a list. If no results are found, return the full list.

        :param id_list: A list of identifiers
        :param search: The search query. This can include regex, but `*` will be replaced with
                       a generic '.*' match all operator.
        :returns: Tuple with:
                      - A list of matching ids if the search is successful. If no ID matches the
                        query, return the full list passed into it.
                      - The true amount of matching results, even if the no results were found.
        """
        if len(id_list) == 0:
            return [], 0  # No need to search an empty list.

        re_search = re.compile(search.upper().replace("*", ".*"))
        search_results = [_id for _id in id_list if re_search.match(_id) is not None]
        if len(search_results) > 0:
            id_list = search_results

        return id_list, len(search_results)

    def _id_search_window(
        self, id_list, search="", id_type="PDB", enable_dp_choice=True
    ):
        """Opens a YASARA window containing a search field and a list. The list will contain the
        PDBs given as the first argument. Additionally, if `enable_dp_choice` is True (default)
        two checkboxes are available for the user to denote whether they would like to download the
        protein structure, the drugs (ligands) or both.

        :param pdbs: A list of IDs that can be selected by the user.
        :param search: A string that will be used as a search query. Optional.
        :returns: A tuple containing:
                  - A string representing the button that was pressed ('OK' or 'Search'),
                  - The search term entered by the user, string
                  - A list of selected ids.
                  (if enable_dp_choice is True:)
                  - A boolean, indicating if drugs should be downloaded.
                  - A boolean, indicating if protein structure should be downloaded.
                  - A boolean, indicating if the _full_ structure should be loaded.
        """

        if search == "":
            list_text = "All available %ss." % id_type
            len_results = len(id_list)
        else:
            id_list, len_results = self._filter_id_list(id_list, search)
            list_text = "%s results for query '%s'" % (len_results, search)

        if len_results == 0:
            id_list, len_results = ["No results"], 1

        if enable_dp_choice:
            results = yasara.ShowWin(
                "Custom",
                "Search %s files" % id_type,
                600,
                400,
                "TextInput",
                20,
                48,
                "Search a %s ID, you can use * as a wildcard." % id_type,
                488,
                100,
                "Button",
                548,
                68,
                "Search",
                "Button",
                550,
                348,
                "_O_ K",
                "Checkbox",
                20,
                348,
                "Compounds",
                "yes",
                "Checkbox",
                180,
                348,
                "Protein",
                "yes",
                "Checkbox",
                310,
                348,
                "Full structure",
                "no",
                # Place the list last to make unpacking easy.
                "List",
                20,
                112,
                list_text,
                560,
                205,
                "yes",
                len_results,
                *id_list
            )
        else:
            results = yasara.ShowWin(
                "Custom",
                "Search %s files" % id_type,
                600,
                400,
                "TextInput",
                20,
                48,
                "Search a %s ID, you can use * as a wildcard." % id_type,
                488,
                100,
                "Button",
                548,
                68,
                "Search",
                "Button",
                550,
                348,
                "_O_ K",
                "List",
                20,
                112,
                list_text,
                560,
                205,
                "yes",
                len_results,
                *id_list
            )

        try:
            button_pressed = results[0]
        except IndexError:
            raise RuntimeWarning("You need to provide input to continue.")

        with Console(yasara):
            print(results)

        if enable_dp_choice:
            search = results[1]
            drugs = bool(results[2])
            protein = bool(results[3])
            full = bool(results[4])
            pdbids = results[6:]

            return button_pressed, search, pdbids, drugs, protein, full
        else:
            search = results[1]
            pdbids = results[3:]
            return button_pressed, search, pdbids

    @staticmethod
    def _switch_to_hud():
        """By default, when yasara.Print() is called, the console is addressed. The console is
        transient and used for more than just displaying scene information, so the HUD is best used
        when trying to show information about the current scene. This function clears the HUD,
        removing all its current contents, redirects printing to it and ensures the cursor starts in
        the top left. This makes sure the HUD is in the proper state to start reporting information.

        After calling this function, information can be displayed in the HUD using either the
        yasara.Print() method, or preferably using `print_to_hud`, which has some special features
        that make HUD interaction easier.

        When you want to restore the default scene information after modifying the HUD, you can
        call `_restore_hud`. This is mainly useful if the information you are providing is of a
        temporary nature.
        """
        yasara.PrintHUD()
        # YASARA will fail to execute FillRect in console mode, stopping all tests.
        if not hasattr(yasara.storage, "running_in_console_mode"):
            yasara.FillRect()
        yasara.PosText(1, 1)

    @staticmethod
    def print_to_hud(text, continue_on_the_same_line=False, ignore_xml_length=True):
        """Prints a line in YASARA, guarding against text overflows. While this uses the standard
        yasara.Print() method and can, as such be used to print to the console, this function is
        best called after _switch_to_hud(). The YASARA HUD is a good way to display information
        about the current scene. Using it, however, means you have to take into account several
        limitations: The HUD can only contain a limited amount of characters, when text overflows
        the HUD boundaries it creates an error state and the HUD allows for certain types of XML to
        be introduced.

        To make displaying information in the HUD easier, this method prevents common pitfalls that
        the limitations cause. Text is automatically wrapped and a newline character is appended to
        the string by defaults. XML groups are ignored in the line count for content to remain
        consistently wrapped. This makes `print_to_hud` a safer alternative to yasara.Print() when
        interacting with the HUD.

        :param text: The text you want to print to the hud. Don't worry about newlines, they are
                     inserted automatically.
        :param continue_on_the_same_line: Optional, defaults to False. If this is set to True, there
                                          will be no newline at the end of this string, allowing you
                                          to continue on the same line with the next print. You may
                                          encounter text overflows if the text you print after it is
                                          too long.
        :param ignore_xml_length: Defaults to True, XML tags will be not not be taken into account
                                  when wrapping text. Since they are invisible, wrapping them may
                                  yield errors or inconsistent newlines.
        :returns: The line of text as it was printed in YASARA.
        """
        max_width = 40  # Normal text only fits 40 characters.
        xml_tags_in_text = []

        xml_re = re.compile("<.+?>")

        if ignore_xml_length:
            for xml_tag in xml_re.finditer(text):
                xml_tags_in_text.append(
                    {
                        "tag": xml_tag.group(0),
                        "start": xml_tag.start(),
                    }
                )

            text = xml_re.sub("", text)  # Remove the xml tags we found from the text.

        text = textwrap.fill(text, max_width)
        if not continue_on_the_same_line:
            text += "\n"

        if ignore_xml_length:  # Restore the xml tags
            for xml_tag in xml_tags_in_text:
                text = (
                    text[: xml_tag["start"]] + xml_tag["tag"] + text[xml_tag["start"] :]
                )

        yasara.Print(text)
        return text

    @staticmethod
    def create_3dm_hud_link(
        text, position, sfamid=None, familyid=None, subsetid=None, numschemeid=-1
    ):
        """Create a link to a 3DM amino acid page that can be used in the HUD.

        :param text: The text that can be clicked.
        :param position: The alignment position.
        :param sfamid: The Superfamily ID you'd like to refer to.
        :param familyid: The ID of the family you are referring to. This will default to 1.
        :param subsetid: The ID of the subset you are referring to. This will default to 1.
        :numschemeid: The numbering scheme ID. Defaults to -1.
        :returns: A string containing an XML action for showing the 3DM url.
        """
        formatted_url = ThreeDMWebsite().create_3dm_residue_link(
            position, sfamid, familyid, subsetid, numschemeid
        )

        url_text = "<action=ShowURL %s>%s</action>" % (formatted_url, text)
        return url_text

    @staticmethod
    def _restore_hud():
        """Restores the YASARA HUD to the normal SCENE CONTENT window."""
        yasara.HUD("Obj", "On", "Normal")

    @staticmethod
    def go_to_3dm_website():
        """Directs the user to the 3DM website using the browser set as the default in YASARA."""
        yasara.ShowURL(ThreeDMWebsite().base_url)

    @staticmethod
    def go_to_help_page():
        """Directs the user to the help page using the browser set as the default in YASARA."""
        # This is a public URL specifically for YASARA
        yasara.ShowURL("https://plugin.bio-prodict.nl/yasara/yasara_installation.html")

    @staticmethod
    def show_about():
        """Shows the user a simple about window."""
        args = ["Custom", "About Bio-Prodict 3DM", 400, 210]

        latest_released_version = Updates.get_latest_stable_version("yasara")
        if VERSION.split(".") < latest_released_version.split("."):
            version_text = "A new version is available: %s" % latest_released_version
        elif VERSION.split(".") > latest_released_version.split("."):
            version_text = "You are using a pre-release version."
        else:
            version_text = "You are up to date."

        current_year = date.today().year
        args += [
            "TextCen",
            200,
            48,
            "3DM YASARA Plugin v%s" % VERSION,
            "TextCen",
            200,
            70,
            "%s, Bio-Prodict B.V." % current_year,
            "TextCen",
            200,
            92,
            "Support: support@bio-prodict.nl",
            "TextCen",
            200,
            114,
            "Website: www.bio-prodict.nl",
            "TextCen",
            200,
            136,
            version_text,
            "Button",
            80,
            166,
            "Go to website",
            "Button",
            350,
            166,
            "OK",
        ]

        about_response = yasara.ShowWin(*args)

        if "Gotowebsite" in about_response:
            yasara.ShowURL("www.bio-prodict.nl")

    @staticmethod
    def show_update_notification(new_version):
        """Shows a window notifying the user that a new version is available."""
        args = [
            "Custom",
            "Update available",
            550,
            210,
            "TextCen",
            275,
            48,
            "A new version of the 3DM plugin is available.",
            "TextCen",
            275,
            70,
            "You are on v%s, v%s is now available for download."
            % (VERSION, new_version),
            "TextCen",
            275,
            114,
            "Please download the latest YASARA to update the plugin.",
            "Button",
            120,
            160,
            "Go to YASARA website",
            "Button",
            500,
            160,
            "Close",
        ]
        response = yasara.ShowWin(*args)

        if "GotoYASARAwebsite" in response:
            yasara.ShowURL("http://yasara.org/update.htm")

    @staticmethod
    def show_proxy_window(message=None):
        """Shows the window that allows the user to submit proxy information.

        :param message: A message displayed to the user.
        """

        args = [
            "Custom",
            "Set proxy information",
            600,
            400,
        ]
        proxies = Proxies.get_proxies()
        if message is not None:
            if isinstance(message, list):
                for i, line in enumerate(message):
                    args += ["TextCen", 300, (48 + 30 * i), line]
            else:
                args += ["TextCen", 300, 48, str(message)]

        http_proxy_string = "HTTP Proxy"
        https_proxy_string = "HTTPS Proxy"

        if "http" in proxies:
            http_proxy_string += " (currently '%s')" % proxies["http"]
        if "https" in proxies:
            https_proxy_string += " (currently '%s')" % proxies["https"]

        args += [
            "TextInput",
            20,
            120,
            http_proxy_string,
            560,
            200,
            "TextInput",
            20,
            190,
            https_proxy_string,
            560,
            200,
            "Button",
            70,
            348,
            "Test proxy",
            "Button",
            450,
            348,
            "_C_ancel",
            "Button",
            550,
            348,
            "_O_K",
        ]

        results = yasara.ShowWin(*args)
        return results

    @staticmethod
    def get_proxies_dict_from_results(results):
        """Converts the output list from the proxies window into a valid proxies dictionary."""
        http_proxy = results[1].strip()
        https_proxy = results[2].strip()
        if https_proxy == "":
            https_proxy = http_proxy

        return {"http": http_proxy, "https": https_proxy}

    def set_proxy_information(self):
        """Allows the user to set and test proxy information."""
        message = None
        while 1:
            proxy_window_response = self.show_proxy_window(message)
            if len(proxy_window_response) < 2 or proxy_window_response[0] == "Cancel":
                yasara.ShowMessage("Proxy settings unchanged.")
                break

            proxies = self.get_proxies_dict_from_results(proxy_window_response)
            if proxy_window_response[0] == "Testproxy":
                try:
                    Proxies.test_proxies(proxies)
                    message = "Proxy settings valid."
                except Exception as e:  # pylint: disable=W0703
                    yasara.Print(str(e))
                    message = [
                        "Could not connect with these proxy settings.",
                        "Check the console for more details.",
                    ]
                    continue
            if proxy_window_response[0] == "OK":
                yasara.ShowMessage("Proxy settings applied.")
                Proxies.dump_proxies_to_file(proxies)
                break

    def pick_system(self):
        """Shows the user a dialog to let them pick a superfamily, returns the SFAMID."""
        access_api = self._get_api("access")
        systems = CompositeUtilities.load_accessible_systems(access_api)
        pretty_system_names = sorted(systems)  # Sorted makes systems easier to find

        picked_pretty_system_name = yasara.ShowWin(
            "List",
            "Select 3DM system",
            "No",
            "Select the 3DM system you would like to access.",
            *pretty_system_names
        )

        try:
            picked_pretty_system_name = picked_pretty_system_name[1]
        except IndexError:
            yasara.ShowMessage("No 3DM system was selected. Exiting plugin.")
            yasara.plugin.end()

        # Get the real SFAMID
        picked_sfamid = systems[picked_pretty_system_name]
        self.storage.system = picked_sfamid  # Store the sfamid
        self.storage.system_nice_name = picked_pretty_system_name

        self.storage.family = 1  # Set the family ID to the default (1)
        self.storage.family_nice_name = "Superfamily 3DM"

        self.storage.subset = 1  # Set the subset ID to the default (1)
        self.storage.all_subsets = CompositeUtilities.load_alignment_subsets(
            self, sfamid=picked_sfamid
        )

        return picked_sfamid

    def pick_family(self):
        """Shows the user a dialog to let them pick a family, returns the family id. This is not
        the same as a superfamily or a subset."""
        try:
            families = CompositeUtilities.load_alignment_families(self)
        except (AttributeError, RuntimeWarning):
            # Ask the user to provide a system
            sfamid = self.pick_system()
            families = CompositeUtilities.load_alignment_families(self, sfamid=sfamid)

        families_inverted = {}
        for pretty_name, famid in families.items():
            n_key = CompositeUtilities.create_formatted_key(famid, pretty_name)
            n_key = n_key.replace("_", "-")  # Remove _ because it cause YASARA to crash
            families_inverted[n_key] = famid

        pretty_family_names = sorted(families_inverted.keys())

        picked_pretty_family_name = yasara.ShowWin(
            "List",
            "Select alignment",
            "No",
            "Select the alignment you would like to access.",
            *pretty_family_names
        )

        try:
            picked_pretty_family_name = picked_pretty_family_name[1]
        except IndexError:
            yasara.ShowMessage("No alignment was selected. Exiting plugin.")
            yasara.plugin.end()

        # Get the real FAMID
        picked_famid = families_inverted[picked_pretty_family_name]
        try:
            different_fam = CompositeUtilities.get_familyid(self) != picked_famid
        except RuntimeWarning:
            different_fam = False

        if different_fam:
            yasara.ShowMessage(
                "Warning: you just picked a new family. Objects are being renumbered to make sure"
                + " that the numbering is correct. This may take some time. Your numbering scheme"
                + "types will be retained."
            )
            self._numbering_refresh(new_family_id=picked_famid)
            yasara.ShowMessage("You are now using family: %s" % (picked_famid))

        self.storage.family = picked_famid
        self.storage.family_nice_name = picked_pretty_family_name
        return picked_famid

    def pick_subset(self):
        """Shows the user a dialog to let them pick a subset, returns the subset id. This is not
        the same as a superfamily, or a family."""
        subsets = self.storage.all_subsets

        subsets_inverted = {}
        for pretty_name, sid in subsets.items():
            n_key = CompositeUtilities.create_formatted_key(sid, pretty_name)
            subsets_inverted[n_key] = sid

        pretty_subset_names = sorted(subsets_inverted.keys())

        picked_pretty_subset_name = yasara.ShowWin(
            "List",
            "Select subset",
            "No",
            "Select the subset you would like to access.",
            *pretty_subset_names
        )

        try:
            picked_pretty_subset_name = picked_pretty_subset_name[1]
        except IndexError:
            yasara.ShowMessage("No subset was selected. Exiting plugin.")
            yasara.plugin.end()

        # Get the real FAMID
        picked_subsetid = subsets_inverted[picked_pretty_subset_name]
        try:
            different_subset = CompositeUtilities.get_subset_id(self) != picked_subsetid
        except RuntimeWarning:
            different_subset = False

        if different_subset:
            yasara.ShowMessage(
                "Warning: you just picked a new subset. Objects are being renumbered to make sure"
                + " that the numbering is correct. This may take some time. Your numbering scheme"
                + "types will be retained."
            )
            self._numbering_refresh(new_subset_id=picked_subsetid)
            yasara.ShowMessage(
                "You are now using subset: %s" % (picked_pretty_subset_name)
            )

        self.storage.subset = int(picked_subsetid)  # Store the famid
        return picked_subsetid

    @staticmethod
    def remove_invalid_chains_from_object(object_id):
        """Removes all the chains from an object that do not belong in the visualization according
        to the object identifier (ie. 1A28B, where B is the chain specifier). Will re-center the
        camera on the now smaller object to enhance user experience.

        :param object_id: A valid object identifier from the scene, 5 character string. If the id
                          is too short, this method will fail with a warning."""
        try:
            valid_chain = object_id[4]
        except IndexError:
            with Console(yasara):
                print(
                    "Warning: Could not remove chains from object without chain specifier in ID."
                )
                return

        yasara.DelRes("Mol !%s in Obj %s" % (valid_chain, object_id))
        yasara.CenterObj(object_id)

    def load_from_3dm(self):
        """Allows the user to pick a PDB from 3DM."""
        self._switch_to_hud()
        yasara.Print("Loading accessible 3DM systems...\n")
        try:
            # Try to get the stored SFAMID first. This saves the user a screen.
            picked_sfamid = CompositeUtilities.get_sfamid(self)
        except (AttributeError, RuntimeWarning):
            picked_sfamid = self.pick_system()

        family = CompositeUtilities.get_familyid(self)

        self.print_to_hud("Loading PDBs...")
        api = self._get_api()
        pdbs = CompositeUtilities.load_pdbs_for_system(self, picked_sfamid, api)

        self._restore_hud()
        button_pressed, search, pdbids, drugs, protein, full = self._id_search_window(
            pdbs
        )
        while button_pressed == "Search":
            # Re-open the PDB Search window until the user has selected an ID
            (
                button_pressed,
                search,
                pdbids,
                drugs,
                protein,
                full,
            ) = self._id_search_window(pdbs, search)

        if not any([full, drugs, protein]):
            raise RuntimeWarning("You selected neither drugs nor proteins to download.")

        specifiers = {
            (True, True, False): "full",  # Whenever full is True, full is used.
            (True, False, True): "full",
            (True, True, True): "full",
            (True, False, False): "full",
            (False, True, False): "hetatm",  # When drugs is True
            (False, False, True): "amino",  # When protein is True
        }
        specifier = specifiers.get((full, drugs, protein), "chain")

        valid_objids = []

        for pdbid in pdbids:
            obj_name = self.load_pdb(
                pdbid, picked_sfamid, specifier, auto_renumber=False
            )
            self.print_to_hud(
                "Successfully loaded PDB %s with obj name %s." % (pdbid, obj_name)
            )

            # Remove invalid chains
            if not drugs or not protein:
                self.remove_invalid_chains_from_object(obj_name)

            valid_objids.append(obj_name)

        self.transfer_pdb_coordinates(valid_objids)

        if protein or full:  # We don't need to renumber when only drugs are loaded.
            for obj_id in valid_objids:
                self.print_to_hud("Renumbering %s..." % obj_id)
                self.renumber_pdb(obj_id, picked_sfamid, family)

            self.print_to_hud("Done renumbering.")

        self._restore_hud()

    def load_model_from_3dm(self):
        """Allows the user to pick a model from 3DM. This is distinct from the structure
        selection functionality."""
        self._switch_to_hud()
        yasara.Print("Loading accessible systems...\n")
        try:
            # Try to get the stored SFAMID first. This saves the user a screen.
            picked_sfamid = CompositeUtilities.get_sfamid(self)
        except (AttributeError, RuntimeWarning):
            picked_sfamid = self.pick_system()

        self.print_to_hud("Loading models...")
        api = self._get_api("persistence3dm")

        models = CompositeUtilities.load_models_for_system(
            picked_sfamid, api, CompositeUtilities.get_familyid(self)
        )

        self._restore_hud()
        button_pressed, search, model_ids = self._id_search_window(
            models, enable_dp_choice=False, id_type="model"
        )
        while button_pressed == "Search":
            # Re-open the Model Search window until the user has selected an ID
            button_pressed, search, model_ids = self._id_search_window(
                models, search, enable_dp_choice=False, id_type="model"
            )

        valid_objids = []

        for model_id in model_ids:
            try:
                CompositeUtilities.load_model(
                    self, model_id, picked_sfamid, CompositeUtilities.get_familyid(self)
                )
                valid_objids.append(model_id)
                self.print_to_hud("Successfully loaded model %s" % model_id)
            except ValueError as exception:
                self.print_to_hud(
                    "Could not load model %s. Is this an alias?" % model_id
                )
                with Console(yasara):
                    print(exception)

        self.transfer_pdb_coordinates(valid_objids)
        self._restore_hud()

    def show_load_basket_dialog(self):
        """Shows a dialog to the user that allows them to pick a basket. Returns the ID of the
        basket the user picked. Ends the plugin if the user did not pick a basket."""

        baskets = CompositeUtilities.load_available_hotspot_baskets(self)
        basket_list = ["%s: %s" % (basket.id, basket.label) for basket in baskets]
        if len(basket_list) == 0:
            raise RuntimeWarning("No baskets exist for this 3DM system.")

        basket = yasara.ShowWin(
            "List",
            "Hotspot baskets",
            "No",
            "Select a basket: " "(Some baskets may exist in different families)",
            *basket_list
        )

        try:
            basket_id = int(basket[1].split(":")[0])
        except IndexError:
            yasara.ShowMessage("No basket was selected. Exiting plugin.")
            yasara.plugin.end()

        return basket_id

    @staticmethod
    def show_basket_items_dialog(basket):
        """Shows the basket items dialog, allowing the user to select one or more items to display
        from the basket.

        :param basket: A Basket instance.
        :returns: A list of positions selected to show by the user.
        """
        items_list = [
            "%s: %s" % (item.get("value"), item.get("label")) for item in basket.items
        ]

        if len(items_list) == 0:
            yasara.ShowMessage("The basket you selected was empty: %s." % basket.label)
            return

        positions = yasara.ShowWin(
            "Custom",
            "Select positions from basket",
            600,
            400,
            "Button",
            550,
            69,
            "OK",
            "Button",
            550,
            114,
            "All",
            "List",
            20,
            48,
            "Select positions from basket",
            485,
            310,
            "yes",
            len(basket.items),
            *items_list
        )

        button_pressed = positions[0]
        n_positions_selected = positions[1]
        selected_positions = []
        if button_pressed != "All" and n_positions_selected > 0:
            selected_positions = [pos.split(":")[0] for pos in positions[2:]]
        else:
            selected_positions = [str(pos.split(":")[0]) for pos in items_list]

        return selected_positions

    def load_hotspot_basket(self):
        """Loads a hotspot basket into the viewer based on user input."""
        try:
            self.check_scene_objects()
        except AttributeError:
            raise RuntimeWarning(
                "You need to load something from 3DM first, or pick a 3DM system."
            )

        api = self._get_baskets_api()
        basket_id = self.show_load_basket_dialog()
        selected_basket = api.load_basket(basket_id)
        self._store_object("current_basket", selected_basket)

        if len(selected_basket.items) > 0:
            selected_positions = self.show_basket_items_dialog(selected_basket)

            # Everything from here on takes place in a new view
            self.duplicate_view("Hotspots - %s" % selected_basket.label)

            values = [1] * len(selected_positions)  # Create a solid coloring base.
            with Console(yasara):
                print(selected_positions)

            (
                selectors,
                renumbered_residues,
                new_properties,
            ) = CompositeUtilities.convert_residues_to_selections(
                self, selected_positions, values, objects=self.name_object("All")
            )

            if len(selectors) == 0:
                raise RuntimeError(
                    "There was an error collecting the positions in this basket."
                )

            print(
                [self.structure_identifier(sel.yasara(), "trunc") for sel in selectors]
            )

            self.show_and_label_residues(selectors, renumbered_residues)
            CompositeUtilities.color_residues_by_property(
                self, selectors, "blue", new_properties
            )
        else:
            yasara.ShowMessage("Loaded empty basket %s." % selected_basket.label)
            # Everything from here on takes place in a new view
            self.duplicate_view(name="Hotspots - %s" % selected_basket.label)

    def create_new_basket(self):
        """Creates a new basket using user input."""
        basket_api = self._get_baskets_api()
        name = ""
        while len(name) < 4:
            name = yasara.ShowWin(
                "TextInput",
                "Basket name",
                "Please enter a name for the basket. It must be at least 4 characters.",
                "Basket name",
            )[0]

        try:
            sfamid = CompositeUtilities.get_sfamid(self)
        except AttributeError:
            sfamid = self.pick_system()

        # If a system has been picked, the family is also available.
        familyid = CompositeUtilities.get_familyid(self)

        basket = basket_api.new_basket(name, sfamid=sfamid, familyid=familyid)
        self._store_object("current_basket", basket)
        yasara.ShowMessage(
            "Successfully created basket '%s', %s" % (name, repr(basket))
        )

    def delete_hotspot_basket(self):
        """Allows the user to select a basket, that will then be deleted."""
        basket_api = self._get_baskets_api()
        basket_id = self.show_load_basket_dialog()
        basket_api.delete_basket_native(basket_id)
        yasara.ShowMessage("Succesfully removed basket '%s'." % basket_id)

    def delete_basket_item(self, positions=None):
        """Allows the user to select a basket, from which positions can then be deleted."""
        if positions is None:
            positions = []
        basket_api = self._get_baskets_api()
        basket_id = self.show_load_basket_dialog()
        basket = basket_api.load_basket(basket_id)
        positions = self.show_basket_items_dialog(basket)
        for position in positions:
            basket.remove_item(where={"value": position})
        basket.save()
        yasara.ShowMessage(
            "Succesfully removed positions from basket '%s'." % basket_id
        )

    @staticmethod
    def get_selected_residues():
        """This method can be called to retrieve the currently selected residue. This supports
        both singular and multi-selections using the Select dialog or selection tools.

        Returns a list of dictionaries containing the residue object from the yasara selection, the
        residue number and the object (PDB) this residue belongs to."""

        if len(yasara.selection[0].residue) > 0:
            selected_residues = [yasara.selection[0].residue[0]]
        elif len(yasara.selection[0].atom) > 0:
            selected_residues = [atom.residue for atom in yasara.selection[0].atom]
        else:
            return None

        residues = {}
        for selected_residue in selected_residues:
            residues[selected_residue.number.inpdb] = {
                "residue": selected_residue,
                "number": selected_residue.number.inpdb,
                "pdb": selected_residue.object.name,
            }

        return list(residues.values())

    def add_residue_to_basket(
        self, basket, residue=None, description=None  # pylint: disable=R0912
    ):
        """Adds residues to a basket. Takes the residues to add from the current YASARA selection,
        be it an Atom or a Residue. This will also convert the residue numbering to the alignment
        numbering scheme, regardless of the scheme used for the selected object.

        :param basket: The Basket object where the residue must be added.
        :param residue: (Optional) The residue number that needs to be added. If this is not
                        specified, the currently selected residue is used.
        :param description: (Optional) The reason why this residue was selected. If this is not
                            specified, the user is prompted for a description."""

        if basket is None:
            yasara.ShowMessage("You need to load or create a basket first.")
            yasara.plugin.end()

        if residue is None:
            residues = self.get_selected_residues()
        else:
            residues = [
                {
                    "pdb": self.name_object("all")[0],
                    "number": residue,
                }
            ]

        pdb = residues[0]["pdb"]  # We can only support one PDB at a time.

        # This prevents errors when working with imported scenes.
        for suffix in ["chain", "_prot"]:
            if pdb.endswith(suffix):
                pdb = pdb.replace(suffix, "")

        if description is None:
            description = yasara.ShowWin(
                "TextInput",
                "Position comment",
                "Describe the reason you added this position (optional).",
                "Description",
            )[0]

        for res in residues:
            new_number = CompositeUtilities.convert_residue_number_from_scene_to_align(
                self, res["pdb"], res["number"]
            )

            # Accept both numbers and strings
            if str(new_number).startswith("-"):
                raise RuntimeWarning(
                    "Could not add position %s to basket, this position exists outside of an "
                    "aligned core. Canceling all additions to this basket." % new_number
                )
            basket.add_residue_position(position=new_number, comment=description)

        self.show_and_label_residues(
            [Selection(numbers=res["number"]) for res in residues],
            [res["number"] for res in residues],
        )

        try:
            basket.save()
            if len(residues) == 1:
                yasara.ShowMessage(
                    "Position %s sucessfully added to basket %s"
                    % (residues[0]["number"], basket.id)
                )
            else:
                yasara.ShowMessage(
                    "Positions sucessfully added to basket %s" % (basket.id)
                )
            self._store_object("current_basket", basket)
        except AssertionError as error:
            yasara.ShowMessage(
                "Couldn't add this position. Is it already in your basket?"
            )
            with Console(yasara):
                print(error)

    def add_residue_to_current_basket(self):
        """Finds the current basket and uses the current selection to add a residue to this
        basket."""
        basket = self._get_stored_object("current_basket")
        self.add_residue_to_basket(basket)

    def add_residue_to_specific_basket(self):
        """Allows the user to pick a basket where they will store the selected residue."""
        api = self._get_baskets_api()
        basket_id = self.show_load_basket_dialog()
        basket = api.load_basket(basket_id)
        self.add_residue_to_basket(basket)

    @staticmethod
    def show_literature_hotspots_search_window():
        """Prompts the user to enter custom keywords for literature hotspots. Returns the words
        in a list."""
        terms = yasara.ShowWin(
            "Custom",
            "Literature Hotspots",
            600,
            250,
            "TextInput",
            20,
            48,
            "Search for literature using keywords: (Ex. activity)",
            488,
            100,
            "Text",
            20,
            140,
            "Minimum number of references:",
            "NumberInput",
            480,
            110,
            " ",
            2,
            1,
            100,
            "Text",
            20,
            200,
            "Maximum amount of results:",
            "NumberInput",
            480,
            170,
            " ",
            25,
            1,
            100,
            "Button",
            548,
            68,
            "Search",
        )
        try:
            return terms[0].split(" "), int(terms[1]), int(terms[2])
        except IndexError:
            raise RuntimeWarning("No values provided for Literature Hotspots.")

    def show_literature_hotspots(self, keywords=None, cutoff=10, limit=25):
        """Shows literature hotspots to the user."""
        try:
            self.check_scene_objects()
        except AttributeError:
            raise RuntimeWarning(
                "You need to load something from 3DM first, or pick a 3DM system."
            )

        if keywords is None:
            keywords, cutoff, limit = self.show_literature_hotspots_search_window()

        mutations_index, articles = CompositeUtilities.get_literature_hotspot_data(
            self, keywords, cutoff=cutoff, limit=limit
        )
        try:
            positions, values = zip(*mutations_index)
        except ValueError:
            raise RuntimeWarning("Keywords '%s' did not yield any results." % keywords)

        self.duplicate_view("Literature Hotspots (%s)" % " ".join(keywords))

        self._switch_to_hud()
        self.print_system_details_to_hud("LITERATURE HOTSPOTS")

        self.print_to_hud("")  # Empty line

        top_n = 20

        self.print_to_hud("POSITION    NUMBER OF MUTATIONS")
        self.print_to_hud("-" * 40)

        (
            selectors,
            renumbered_residues,
            new_properties,
        ) = CompositeUtilities.convert_residues_to_selections(
            self, positions, values, objects=self.name_object("All")
        )

        converted_positions = {}
        for p, (pos, n_mutations) in enumerate(
            mutations_index[:top_n]
        ):  # pylint: disable=W0612
            pos_str = str(renumbered_residues[p]).ljust(8)
            converted_positions[pos] = str(renumbered_residues[p])

            literature_link = ThreeDMWebsite().create_3dm_literature_link(
                pos,
                keywords,
                CompositeUtilities.get_sfamid(self),
                CompositeUtilities.get_familyid(self),
                CompositeUtilities.get_subset_id(self),
            )

            n_mutations_link = "<action=ShowURL %s>%s</action>" % (
                literature_link,
                n_mutations,
            )

            self.print_to_hud("%s    %s" % (pos_str, n_mutations_link))

        self.print_to_hud("")
        self.print_to_hud(
            "Only the top %i are shown here. Click on show literature for more." % top_n
        )

        action = (
            "SavePLIAtom None,Request=show_literature_hotspot_articles"
            "|RunPlugin bioprodict_3dm_v2.py"
        )

        yasara.ShowButton("Show Literature", x=70, y=400, action=action)

        yasara.ColorRes("all", "gray")  # Color everything grey.

        self.show_and_label_residues(selectors, renumbered_residues)
        CompositeUtilities.color_residues_by_property(
            self, selectors, "red", new_properties
        )

        for article in articles:
            try:
                article["position"] = converted_positions[article["position"]]
            except KeyError:
                pass

        self.storage.lh_articles = articles

    def show_literature_hotspot_articles(self):
        """Shows the user a list of articles that correspond to the latest search in 3DM. This
        dialog allows the user to select articles and subsequently open them in their browser for
        reading."""
        try:
            articles = self.storage.lh_articles
        except AttributeError:
            raise RuntimeWarning(
                "You need to have looked up articles with Show Literature Hotspots."
            )

        list_text = "%s articles found. (Mutation/Position - Organism - PubMed)" % len(
            articles
        )

        parsed_articles = [
            "%s/%s - %s - %s - %s"
            % (
                article["mutation"],
                str(article["position"]),
                article["organism_id"],
                (
                    article["title"].encode("ascii", errors="ignore")
                    if article["title"] is not None
                    else "Unknown title."
                ),
                article["pubmed_id"],
            )
            for article in articles
        ]

        results = yasara.ShowWin(
            "Custom",
            "Articles",
            600,
            400,
            "Button",
            518,
            350,
            "Open selected",
            "Button",
            52,
            350,
            "Close",
            # Place the list last to make unpacking easy.
            "List",
            20,
            48,
            list_text,
            560,
            275,
            "yes",
            len(parsed_articles),
            *parsed_articles
        )
        if len(results) == 0:
            # Resort to default close behaviour if the user did not provide input
            results = ["close"]

        base_url = "https://www.ncbi.nlm.nih.gov/pubmed/"
        if results[0] == "Openselected":
            if results[1] > 1:
                ids = ",".join([result.split(" - ")[-1] for result in results[2:]])
            else:
                ids = results[2].split(" - ")[-1]
            content = "?report=abstract"

            yasara.ShowURL(base_url + ids + content)

    def print_system_details_to_hud(self, title, object_name=None):
        """Print details about the current system to the YASARA HUD. Requires that `_switch_to_hud`
        has been called beforehand.

        :param title: A title that is displayed as part of the details header. This title is
                      displayed in upper case.
        :param object_name: Optional, if this is specified the numbering scheme of this object will
                            also be shown.
        """
        self.print_to_hud("")
        self.print_to_hud(title.upper())
        self.print_to_hud("-" * 40)

        nice_sfamid = CompositeUtilities.get_sfamid_nice_name(self)
        nice_famid = CompositeUtilities.get_family_nice_name(self)
        subset_id = CompositeUtilities.get_subset_id(self)

        subset = {v: k for k, v in self.storage.all_subsets.items()}[subset_id]

        if len(nice_sfamid) > 27:
            nice_sfamid = nice_sfamid[:24] + "..."
        if len(nice_famid) > 27:
            nice_famid = nice_famid[:24] + "..."

        self.print_to_hud("Superfamily: %s" % nice_sfamid)
        self.print_to_hud("Family:      %s" % nice_famid)
        self.print_to_hud("Subset:      %s" % subset)

        if object_name is not None:
            try:
                nice_numbering_scheme_name = self.storage.numbering_schemes.get(
                    object_name
                )
            except AttributeError:
                return

            nice_numbering_scheme_name = (
                CompositeUtilities.get_nice_name_for_num_scheme_id(
                    self, nice_numbering_scheme_name
                )
            )
            self.print_to_hud("Numbering:   %s" % nice_numbering_scheme_name)

    def show_amino_acid_distribution(self):
        """Shows the amino acid distribution for the currently selected residue."""
        residues = self.get_selected_residues()
        self.check_scene_objects()

        raw_residue_number = residues[0]["number"]

        residue_number = CompositeUtilities.convert_residue_number_from_scene_to_align(
            self, residues[0]["pdb"], raw_residue_number
        )

        cons_values = CompositeUtilities.get_amino_acid_distribution(
            self, residue_number
        )
        metadata = CompositeUtilities.get_amino_acid_distribution_details(
            self, residue_number
        )

        total = metadata["total"]
        total_gaps = metadata["total_gaps"]

        cons_values_sorted = sorted(
            cons_values.items(), key=lambda x: x[1], reverse=True
        )

        sfamid = CompositeUtilities.get_sfamid(self)
        famid = CompositeUtilities.get_familyid(self)
        subset_id = CompositeUtilities.get_subset_id(self)

        self._switch_to_hud()
        self.print_system_details_to_hud(
            "Amino acid distribution for residue %s" % raw_residue_number,
            object_name=residues[0].get("pdb"),
        )
        self.print_to_hud("Total amount of sequences: %i" % total)
        self.print_to_hud("Total amount of gaps:      %i" % total_gaps)

        self.print_to_hud("")
        self.print_to_hud("AMINO ACID  PERCENTAGE  SEQUENCES")
        self.print_to_hud("-" * 40)
        for key, val in cons_values_sorted:
            if key == "-":
                continue
            abs_val = int((total - total_gaps) * val)
            if val > 0:
                self.print_to_hud(
                    key.ljust(12) + ("%.2f%%" % (val * 100)).ljust(12) + str(abs_val)
                )

        readmore = self.create_3dm_hud_link(
            "Click for more details",
            residue_number,
            sfamid=sfamid,
            subsetid=subset_id,
            familyid=famid,
        )
        self.print_to_hud("")
        self.print_to_hud(readmore)

        self.print_to_hud("-" * 40)

    @staticmethod
    def show_filter_options_window(prop):
        """Shows the user a window allowing them to pick filtering options."""
        # Superfamily property options are defined as follows:
        # nicename: The title of the window, the property name with better punctuation etc.
        # description: Description of what is about to be displayed.
        # cutoff_description: Description of the cutoff mechanic
        # cutoff_default: The default cutoff value.

        superfamily_property_options = {
            "topconserved": {
                "nicename": "Conservation score",
                "description": "conserved in the 3DM system",
                "cutoff_description": "Minimum conservation",
                "cutoff_default": 80,
            },
            "correlatedmutations": {
                "nicename": "Correlations",
                "description": "correlation network",
                "cutoff_description": "Correlation cutoff",
                "cutoff_default": 1,
            },
            "bvalue": {
                "nicename": "B-factor",
                "description": "posititions with a high B-factor",
                "cutoff_description": "Minimum B-factor",
                "cutoff_default": 25,
            },
            "ligandcontacts": {
                "nicename": "ligand contacts",
                "description": "ligand contacts",
                "cutoff_description": "Min. amount of contacts",
                "cutoff_default": 10,
            },
            "mutations": {
                "nicename": "Mutations",
                "description": "positions with high amounts of mutations",
                "cutoff_description": "Min. number of mutations",
                "cutoff_default": 5,
            },
            "rmsd": {
                "nicename": "RMSD",
                "description": "positions with high RMSD",
                "cutoff_description": "Minimum RMSD",
                "cutoff_default": 2,
            },
            "pdbresiduecontactsexternal": {
                "nicename": "External protein-protein interactions",
                "description": "protein contacts outside the structure",
                "cutoff_description": "Min. contact count",
                "cutoff_default": 2,
            },
            "pdbresiduecontactsinternal": {
                "nicename": "Intra-complex protein interactions",
                "description": "protein contacts inside the structure",
                "cutoff_description": "Min. contact count",
                "cutoff_default": 2,
            },
            "dnarnacontacts": {
                "nicename": "DNA/RNA contacts",
                "description": "positions with DNA or RNA contacts",
                "cutoff_description": "Min. contact count",
                "cutoff_default": 10,
            },
            "default": {
                "nicename": prop.capitalize(),
                "description": "positions by %s" % prop,
                "cutoff_description": "Cutoff value",
                "cutoff_default": 1,
            },
        }

        options = superfamily_property_options.get(
            prop, superfamily_property_options["default"]
        )
        try:
            maxvalue, cutoff = yasara.ShowWin(
                "Custom",
                options["nicename"],
                430,
                220,
                "Text",
                20,
                48,
                "Show %s." % options["description"],
                "Text",
                20,
                78,
                "Positions",
                "Text",
                200,
                78,
                options["cutoff_description"],
                "NumberInput",
                20,
                88,
                " ",
                20,
                1,
                200,
                "NumberInput",
                200,
                88,
                " ",
                options["cutoff_default"],
                1,
                100,
                "Button",
                378,
                168,
                "_O_K",
            )
        except ValueError:
            raise RuntimeWarning("User exited input window. Exiting plugin.")

        return maxvalue, cutoff

    def color_molecules_by(self, prop):
        """Colors molecules according to a property from the superfamily data, ie Cores or
        Hydrophobicity.

        :param prop: String, a property of the superfamily data.
        """

        try:
            self.check_scene_objects()
        except AttributeError:
            raise RuntimeWarning(
                "You need to load something from 3DM first, or pick a 3DM system."
            )

        self.check_structures_exist_in_scene()

        data = CompositeUtilities.download_superfamily_data(self)

        if prop == "cores":
            # "Cores isn't really a property, it's actually just position numbers
            position_data = {
                pos["alignmentPosition"]: pos["alignmentPosition"] for pos in data
            }
        else:
            position_data = {pos["alignmentPosition"]: pos[prop] for pos in data}

        self.duplicate_view(prop.capitalize())
        yasara.ColorRes("all", "gray")  # Color everything grey.
        positions = position_data.keys()
        values = list(position_data.values())

        (
            selectors,
            renumbered_residues,
            new_properties,
        ) = CompositeUtilities.convert_residues_to_selections(
            self, positions, values, objects=self.name_object("All")
        )

        del renumbered_residues

        if prop == "cores":
            CompositeUtilities.color_residues_by_property(
                self, selectors, "red", new_properties, loop=True
            )
        elif prop == "hydrophobicity":
            CompositeUtilities.color_residues_by_property(
                self, selectors, "red", new_properties, start_color="green"
            )
        else:
            CompositeUtilities.color_residues_by_property(
                self, selectors, "red", new_properties
            )

    @staticmethod
    def find_obj_name_for_compound_mol(compound_mol_str):
        """Returns the name of the first object that matches of the given compound mol string."""
        for mol, cp_mol in zip(yasara.ListMol("All"), yasara.CompoundMol("All")):
            if cp_mol == compound_mol_str:
                return yasara.NameObj(mol)[0]

    def show_superfamily_data(self, prop, maxvalue=None, cutoff=None):
        """Shows the superfamily data specified by prop by utilizing the positionInfo API. Works
        in a new View."""
        try:
            self.check_scene_objects()
        except AttributeError:
            raise RuntimeWarning(
                "You need to load something from 3DM first, or pick a 3DM system."
            )

        self.check_structures_exist_in_scene()

        data = CompositeUtilities.download_superfamily_data(self)
        if maxvalue is None or cutoff is None:
            maxvalue, cutoff = self.show_filter_options_window(prop)

        position_data = CompositeUtilities.filter_position_data(
            data, prop, maxvalue=maxvalue, cutoff=cutoff
        )
        positions, values = zip(
            *sorted(position_data.items(), key=lambda x: (x[1], [0]), reverse=True)
        )

        self.duplicate_view(name=prop.capitalize())
        yasara.ColorRes("all", "gray")  # Color everything grey.
        yasara.UnlabelRes("all")  # Removes all the labels

        (
            selectors,
            renumbered_residues,
            new_properties,
        ) = CompositeUtilities.convert_residues_to_selections(
            self, positions, values, objects=self.name_object("All")
        )

        if not renumbered_residues:
            n_objects = len(yasara.CompoundMol("All"))
            if not n_objects:
                raise RuntimeWarning("No structures loaded.")
            else:
                raise RuntimeWarning("No valid structures loaded.")

        self.show_in_hud(prop, positions, values, renumbered=renumbered_residues)

        self.show_and_label_residues(selectors, renumbered_residues)
        CompositeUtilities.color_residues_by_property(
            self, selectors, "red", new_properties
        )

    def show_in_hud(self, prop, positions, values, renumbered=None):
        """Shows the values and positions in the HUD for a given property.

        :param prop: The property you are displaying.
        :param positions: A list of integers referring to the alignment positions you are showing.
                          These will be converted to links to their 3DM page.
        :param values: The value of the given property for this position.
        :param renumbered: A dictionary containing adjusted residues numbers (optional)
        """

        sfamid = CompositeUtilities.get_sfamid(self)
        famid = CompositeUtilities.get_familyid(self)
        subset_id = CompositeUtilities.get_subset_id(self)

        self._switch_to_hud()
        self.print_system_details_to_hud("Details for property %s" % prop)

        self.print_to_hud("")
        self.print_to_hud("RANK  POS          VALUE")
        self.print_to_hud("-" * 40)
        for i, (pos, val) in enumerate(zip(positions, values)):
            rank = str(i + 1).rjust(3, " ") + ")"
            if renumbered is not None:
                try:
                    pos_name = renumbered[i]
                except IndexError:
                    raise RuntimeError(
                        "Received invalid renumbering scheme while showing residue properties "
                        "in the HUD."
                    )
            else:
                pos_name = pos

            position_with_url = self.create_3dm_hud_link(
                pos_name, pos, sfamid=sfamid, subsetid=subset_id, familyid=famid
            )
            position = position_with_url + " " * (
                10 - len(str(pos_name))
            )  # Justify position.
            value = str(round(val, 2)).ljust(6, " ")
            self.print_to_hud("  ".join([rank, position, value]))
        self.print_to_hud("-" * 40)

    def show_and_label_residues(self, selectors, residue_numbers):
        """Adds a label and shows the residue of each of the given YASARA selections. The label
        will correspond to the first integer in the selection. If the label does not contain an int,
        it will be skipped.

        :param selectors: A list of Selection objects
        :param residue_numbers: A list of residue numbers that should serve as a label,
        corresponding to the `selectors` list.
        """

        yasara.UnlabelRes("All")  # Remove all previous labels.
        yasara.HideRes("Protein")  # Hide all previous residues.

        for selector, number in zip(selectors, residue_numbers):
            selector = self.structure_identifier(selector, mode="trunc")

            # Get the first integer from the string.
            yasara.LabelRes(selector.yasara(), str(number))
            yasara.ShowRes(selector.yasara())  # Shows the residue atoms

    def _color_residue(self, selector, color):
        """Colors a residue.

        :param selector: Residue selector.
        :param color: A color in hex format or YASARA color format."""
        selector = self.structure_identifier(selector, mode="trunc")
        yasara.ColorRes(selector.yasara(), color)

    def inspect_storage(self):
        """Not exposed in the UI, this function stores the current `storage` object in a temporary
        file and opens it in the default browser. The file is deleted after it is opened. This
        allows a developer to inspect what is currently stored in the `storage` container. This
        function is accessed by entering the following in the YASARA terminal:

        .. code-block:: python

            SavePLIObj All,inspect_storage
            RunPlugin bioprodict_3dm_v2.py

        It is suggested that you install a JSON viewer plugin in order to get a better view of the
        data.

        """
        serializable = {}
        for key, value in self.storage.__dict__.items():
            try:
                json.dumps({key: value})
                serializable[key] = value
            except TypeError:
                serializable[key] = str(value)

        storage_string = json.dumps(serializable, separators=(",", ":"))

        pdbfile = tempfile.NamedTemporaryFile(
            delete=False, suffix="inspect_storage.json", mode="w"
        )
        abs_pdb_path = os.path.abspath(pdbfile.name)
        try:
            pdbfile.write(storage_string)
        except TypeError:
            pdbfile.write(storage_string.encode("utf8"))
        pdbfile.close()  # Close the PDB file so it can be read under Windows.

        if not serializable.get("running_in_console_mode", False):
            yasara.ShowURL("file://%s" % abs_pdb_path)
            time.sleep(5)  # Wait a few seconds before deleting

        yasara.DelFile(abs_pdb_path)
        return storage_string

    @staticmethod
    def show_error_feedback_dialog(traceback):
        """This dialog shows the traceback that will be sent to Bio-Prodict, should the user
        decide to do so.

        :param traceback: The traceback that will be shown to the user.
        """
        traceback_lines = [line for line in traceback.split("\n") if len(line)]

        # Button order is deliberate; the final button is the one that is activated when a line
        # in the list is double clicked by accident. This should cancel by default.
        feedback_response = yasara.ShowWin(
            "Custom",
            "Unexpected error occurred",
            600,
            400,
            "Button",
            550,
            348,
            "Send",
            "Button",
            54,
            348,
            "Cancel",
            "List",
            20,
            48,
            "Click send to inform Bio-Prodict and receive support.",
            560,
            265,
            "no",
            len(traceback_lines),
            *traceback_lines
        )
        return "Send" in feedback_response

    @staticmethod
    def get_version():
        """Returns the version of the plugin and other dependencies."""
        commit_hash_file = os.path.join(os.path.dirname(__file__), "..", "commit.txt")
        commit_hash = "no-commit-hash"
        try:
            with open(commit_hash_file) as commit_hash_handle:
                commit_hash = commit_hash_handle.read().strip()
        except IOError:
            # No commit hash file, there have been no commits or the post-commit hook has not
            # been set up correctly.
            pass

        return "bioprodict-yasara-plugin/%s (%s) python-requests/%s yasara/%s" % (
            VERSION,
            commit_hash,
            requests.__version__,
            yasara.version,
        )
